from abc import ABC, abstractmethod
from typing import Generic, TypeVar


T = TypeVar("T")


class SpectrumEmulator(Generic[T]):
    @property
    @abstractmethod
    def stellar_parameter_names(self) -> T:
        """Get labels of stellar parameters (no geometry-related parameters, e.g. mu)

        Returns:
            T:
        """
        raise NotImplementedError
    
    @abstractmethod
    def to_parameters(self, parameters: T) -> T:
        raise NotImplementedError

    @abstractmethod
    def flux(self, log_wavelengths: T, parameters: T) -> T:
        raise NotImplementedError

    @abstractmethod
    def intensity(self, log_wavelengths: T, mu: float, parameters: T) -> T:
        """Calculate the intensity for given wavelengths and mus

        Args:
            log_wavelengths (T): [log(angstrom)]
            mu (float): cosine of the angle between the star's radius and the line of sight
            spectral_parameters (T): an array of predefined stellar parameters

        Returns:
            T: intensities corresponding to passed wavelengths [erg/cm2/s/angstrom]
        """
        raise NotImplementedError



from abc import ABC
from typing import Optional
import re
import matplotlib.pyplot as plt
from spice.spectrum.utils import wavelengths_to_frequencies
import numpy as np
from importlib import resources as impresources
from . import filter_data

try:
    import jax.numpy as jnp
    from jax.typing import ArrayLike
    import jax
    from jaxtyping import Array, Float

except ImportError:
    raise ImportError("Please install JAX to use SPICE.")


@jax.jit
def _filter_responses(wavelengths: ArrayLike,
                      sample_wavelengths: ArrayLike,
                      sample_responses: ArrayLike) -> ArrayLike:
    return jnp.interp(wavelengths, sample_wavelengths, sample_responses)


class Filter(ABC):
    def __init__(self,
                 transmission_curve: Float[Array, "2 n_samples"],
                 name: Optional[str] = None,
                 non_photonic: bool = False,
                 AB_zeropoint: float = -22.40788262039795,
                 ST_zeropoint: float = -21.10,
                 Vega_zeropoint: Optional[float] = None):
        """A class representing an astronomical filter with its transmission curve and properties.

        The Filter class provides functionality to handle astronomical filters, including their
        transmission curves in both wavelength and frequency space, and methods to apply the
        filter response to spectral data.

        Args:
            transmission_curve (Float[Array, "2 n_samples"]): A 2xN array containing wavelengths
                and corresponding transmission values. First row should be wavelengths in Angstroms,
                second row should be transmission values between 0 and 1.
            ab_zeropoint (float, optional): The AB magnitude system zero point in erg/s/cm^2/Hz.
                Defaults to 3.631e-20 (standard AB system zero point).
            name (Optional[str], optional): Name of the filter. If None, will be derived from class name.
                Defaults to None.
            non_photonic (bool, optional): Whether the filter is non-photonic and a different form of synthetic photometry equation is used. Defaults to False.
            AB_zeropoint (float, optional): The zeropoint of the filter in magnitudes in the AB system. Defaults to -22.407 (-2.5 log10(3.631 * 1e-20 * c [km/s] * 1e5)).
            ST_zeropoint (float, optional): The zeropoint of the filter in magnitudes in the ST system. Defaults to -21.10.
            Vega_zeropoint (float, optional): The zeropoint of the filter in magnitudes in the Vega system. Defaults to None.
        Properties:
            transmission_curve_wavelengths: The transmission curve in wavelength space
            transmission_curve_frequencies: The transmission curve in frequency space
            ab_zeropoint: The AB magnitude system zero point
            name: The name of the filter
            non_photonic: Whether the filter is non-photonic and a different form of synthetic photometry equation is used
        Methods:
            filter_responses_for_wavelengths: Get filter response values for given wavelengths
            filter_responses_for_frequencies: Get filter response values for given frequencies
            plot_filter_responses_for_wavelengths: Plot filter response curve with input spectrum
        """
        self.__transmission_curve = transmission_curve

        tc_freq = wavelengths_to_frequencies(transmission_curve[0])
        tc_freq_mask = jnp.argsort(tc_freq)
        self.__transmission_curve_freq = jnp.vstack([tc_freq[tc_freq_mask],
                                                     transmission_curve[1][tc_freq_mask]])

        self.__name = name or re.sub(r"([A-Z])", r" \1", type(self).__name__).split()
        self.__non_photonic = non_photonic
        self.__AB_zeropoint = AB_zeropoint
        self.__ST_zeropoint = ST_zeropoint
        self.__Vega_zeropoint = Vega_zeropoint
    @property
    def transmission_curve_wavelengths(self) -> ArrayLike:
        return self.__transmission_curve

    @property
    def transmission_curve_frequencies(self) -> ArrayLike:
        return self.__transmission_curve_freq

    @property
    def AB_zeropoint(self) -> float:
        return self.__AB_zeropoint
    
    @property
    def ST_zeropoint(self) -> float:
        return self.__ST_zeropoint
    
    @property
    def Vega_zeropoint(self) -> float:
        return self.__Vega_zeropoint

    @property
    def name(self) -> str:
        return self.__name

    @property
    def non_photonic(self) -> bool:
        return self.__non_photonic
    
    @property
    def zeropoint(self) -> float:
        return self.__zeropoint

    def __str__(self):
        return self.name

    def filter_responses_for_wavelengths(self, wavelengths: Float[Array, "n_wavelengths"]) -> Float[Array, "n_wavelengths"]:
        return _filter_responses(wavelengths, self.transmission_curve_wavelengths[0],
                                 self.transmission_curve_wavelengths[1])

    def filter_responses_for_frequencies(self, wavelengths: Float[Array, "n_wavelengths"]) -> Float[Array, "n_wavelengths"]:
        return _filter_responses(wavelengths, self.transmission_curve_frequencies[0],
                                 self.transmission_curve_frequencies[1])

    def plot_filter_responses_for_wavelengths(self,
                                              wavelengths: Float[Array, "n_wavelengths"],
                                              intensities: Float[Array, "n_wavelengths"],
                                              plot_kwargs: Optional[dict] = None):
        plot_kwargs = plot_kwargs or {}
        if "label" not in plot_kwargs:
            plot_kwargs["label"] = self.name

        responses = self.filter_responses_for_wavelengths(wavelengths)
        plt.plot(wavelengths, intensities, color='black')
        plt.plot(wavelengths, intensities * responses, **plot_kwargs)
        plt.gca().set_xlabel("Wavelength [$\\AA$]")
        plt.gca().set_ylabel("Intensity [$\\frac{erg}{s\\ cm^3}$]")
        plt.legend()

    def plot_filter_responses_for_frequencies(self,
                                              frequencies: Float[Array, "n_wavelengths"],
                                              intensities: Float[Array, "n_wavelengths"],
                                              plot_kwargs: Optional[dict] = None):
        plot_kwargs = plot_kwargs or {}
        if "label" not in plot_kwargs:
            plot_kwargs["label"] = self.name

        responses = self.filter_responses_for_frequencies(frequencies)
        plt.plot(frequencies, intensities, color='black')
        plt.plot(frequencies, intensities * responses, **plot_kwargs)
        plt.gca().set_xlabel("Frequency [Hz]")
        plt.gca().set_ylabel("Intensity [$\\frac{erg}{s\\ cm^3}$]")
        plt.legend()


# Filter information: Bessel & Murphy (2012)

class JohnsonCousinsU(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [3000., 3050., 3100., 3150., 3200., 3250., 3300., 3350., 3400.,
             3450., 3500., 3550., 3600., 3650., 3700., 3750., 3800., 3850.,
             3900., 3950., 4000., 4050., 4100., 4150., 4200.],
            [0.000, 0.019, 0.068, 0.167, 0.278, 0.398, 0.522, 0.636, 0.735,
             0.813, 0.885, 0.940, 0.980, 1.000, 1.000, 0.974, 0.918, 0.802,
             0.590, 0.355, 0.194, 0.107, 0.046, 0.003, 0.000]
        ])
        super().__init__(transmission_curve, name='Johnson-Cousins U', AB_zeropoint=-22.40788262039795-0.771+0.01, ST_zeropoint=-21.10+0.142)


class JohnsonCousinsB(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [3600., 3700., 3800., 3900., 4000., 4100., 4200., 4300., 4400.,
             4500., 4600., 4700., 4800., 4900., 5000., 5100., 5200., 5300.,
             5400., 5500., 5600.],
            [0.000, 0.031, 0.137, 0.584, 0.947, 1.000, 1.000, 0.957, 0.895,
             0.802, 0.682, 0.577, 0.474, 0.369, 0.278, 0.198, 0.125, 0.078,
             0.036, 0.008, 0.000]
        ])
        super().__init__(transmission_curve, name='Johnson-Cousins B', AB_zeropoint=-22.40788262039795+0.138-0.008, ST_zeropoint=-21.10+0.625)


class JohnsonCousinsV(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [4700., 4800., 4900., 5000., 5100., 5200., 5300., 5400., 5500.,
             5600., 5700., 5800., 5900., 6000., 6100., 6200., 6300., 6400.,
             6500., 6600., 6700., 6800., 6900., 7000., 7100., 7200., 7300.,
             7400.],
            [0.000, 0.033, 0.176, 0.485, 0.811, 0.986, 1.000, 0.955, 0.865,
             0.750, 0.656, 0.545, 0.434, 0.334, 0.249, 0.180, 0.124, 0.075,
             0.041, 0.022, 0.014, 0.011, 0.008, 0.006, 0.004, 0.002, 0.001,
             0.000]
        ])
        super().__init__(transmission_curve, name='Johnson-Cousins V', AB_zeropoint=-22.40788262039795+0.023-0.003, ST_zeropoint=-21.10+0.019)


class JohnsonCousinsR(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [5500., 5600., 5700., 5800., 5900., 6000., 6100., 6200., 6300.,
             6400., 6500., 6600., 6700., 6800., 6900., 7000., 7100., 7200.,
             7300., 7400., 7500., 7600., 7700., 7800., 7900., 8000., 8100.,
             8200., 8300., 8400., 8500., 8600., 8700., 8800., 8900., 9000.,
             9100.],
            [0.000, 0.247, 0.780, 0.942, 0.998, 1.000, 0.974, 0.940, 0.901,
             0.859, 0.814, 0.760, 0.713, 0.662, 0.605, 0.551, 0.497, 0.446,
             0.399, 0.350, 0.301, 0.257, 0.215, 0.177, 0.144, 0.116, 0.089,
             0.066, 0.051, 0.039, 0.030, 0.021, 0.014, 0.008, 0.006, 0.003,
             0.000]
        ])
        super().__init__(transmission_curve, name='Johnson-Cousins R', AB_zeropoint=-22.40788262039795-0.16-0.003, ST_zeropoint=-21.10-0.538)


class JohnsonCousinsI(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [7000., 7100., 7200., 7300., 7400., 7500., 7600., 7700., 7800.,
             7900., 8000., 8100., 8200., 8300., 8400., 8500., 8600., 8700.,
             8800., 8900., 9000., 9100., 9200.],
            [0.000, 0.090, 0.356, 0.658, 0.865, 0.960, 1.000, 0.998, 0.985,
             0.973, 0.970, 0.958, 0.932, 0.904, 0.860, 0.810, 0.734, 0.590,
             0.392, 0.203, 0.070, 0.008, 0.000]
        ])
        super().__init__(transmission_curve, name='Johnson-Cousins I', AB_zeropoint=-22.40788262039795-0.402-0.002, ST_zeropoint=-21.10-1.220)


class HipparcosHp(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [3400., 3500., 3600., 3700., 3800., 3900., 4000., 4100., 4200.,
             4300., 4400., 4500., 4600., 4700., 4800., 4900., 5000., 5100.,
             5200., 5300., 5400., 5500., 5600., 5700., 5800., 5900., 6000.,
             6100., 6200., 6300., 6400., 6500., 6600., 6700., 6800., 6900.,
             7000., 7100., 7200., 7300., 7400., 7500., 7600., 7700., 7800.,
             7900., 8000., 8100., 8200., 8300., 8400., 8500., 8600., 8700.,
             8800., 8900.],
            [0.000, 0.041, 0.072, 0.133, 0.199, 0.263, 0.347, 0.423, 0.508,
             0.612, 0.726, 0.813, 0.906, 0.966, 0.992, 1.000, 0.992, 0.978,
             0.951, 0.914, 0.880, 0.840, 0.797, 0.755, 0.712, 0.668, 0.626,
             0.583, 0.542, 0.503, 0.465, 0.429, 0.393, 0.359, 0.326, 0.293,
             0.260, 0.230, 0.202, 0.176, 0.152, 0.130, 0.112, 0.095, 0.081,
             0.068, 0.054, 0.042, 0.032, 0.024, 0.018, 0.014, 0.010, 0.006,
             0.002, 0.000]
        ])
        super().__init__(transmission_curve, name='Hipparcos Hp', AB_zeropoint=-22.40788262039795+0.022+0.008, ST_zeropoint=-21.10+0.074)


class TychoBT(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [3500., 3550., 3600., 3650., 3700., 3750., 3800., 3850., 3900.,
             3950., 4000., 4050., 4100., 4150., 4200., 4250., 4300., 4350.,
             4400., 4450., 4500., 4550., 4600., 4650., 4700., 4750., 4800.,
             4850., 4900., 4950., 5000., 5050.],
            [0.000, 0.015, 0.063, 0.132, 0.220, 0.323, 0.439, 0.556, 0.664,
             0.751, 0.813, 0.853, 0.880, 0.904, 0.931, 0.960, 0.984, 1.000,
             0.969, 0.852, 0.674, 0.479, 0.309, 0.196, 0.131, 0.097, 0.077,
             0.056, 0.035, 0.015, 0.003, 0.000]
        ])
        super().__init__(transmission_curve, name='Tycho BT', AB_zeropoint=-22.40788262039795+0.09+0.01, ST_zeropoint=-21.10+0.672)
        

class TychoVT(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [4550., 4600., 4650., 4700., 4750., 4800., 4850., 4900., 4950.,
             5000., 5050., 5100., 5150., 5200., 5250., 5300., 5350., 5400.,
             5450., 5500., 5550., 5600., 5650., 5700., 5750., 5800., 5850.,
             5900., 5950., 6000., 6050., 6100., 6150., 6200., 6250., 6300.,
             6350., 6400., 6450., 6500., 6550., 6600., 6650., 6700., 6750.],
            [0.000, 0.023, 0.119, 0.308, 0.540, 0.749, 0.882, 0.951, 0.981,
             0.997, 1.000, 0.992, 0.974, 0.946, 0.911, 0.870, 0.827, 0.784,
             0.738, 0.692, 0.645, 0.599, 0.553, 0.504, 0.458, 0.412, 0.368,
             0.324, 0.282, 0.245, 0.209, 0.178, 0.152, 0.129, 0.108, 0.092,
             0.078, 0.066, 0.055, 0.044, 0.036, 0.027, 0.017, 0.008, 0.000]
        ])
        super().__init__(transmission_curve, name='Tycho VT', AB_zeropoint=-22.40788262039795+0.044-0.007, ST_zeropoint=-21.10+0.115)


# # J/A+A/649/A3     Gaia Early Data Release 3 photometric passbands (Riello+, 2021)

class GaiaG(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / 'gaia_edr3_passband.dat')).T
        zp = np.loadtxt((impresources.files(filter_data) / 'gaia_edr3_zeropt.dat'))
        transmission_curve[transmission_curve == 99.99] = 0.
        super().__init__(jnp.array([transmission_curve[0]*10., transmission_curve[1]]), name='Gaia G', AB_zeropoint=zp[1, 0], Vega_zeropoint=zp[0, 0])


class GaiaBP(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / 'gaia_edr3_passband.dat')).T
        zp = np.loadtxt((impresources.files(filter_data) / 'gaia_edr3_zeropt.dat'))
        transmission_curve[transmission_curve == 99.99] = 0.
        super().__init__(jnp.array([transmission_curve[0]*10., transmission_curve[3]]), name='Gaia BP', AB_zeropoint=zp[1, 2], Vega_zeropoint=zp[0, 2])


class GaiaRP(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / 'gaia_edr3_passband.dat')).T
        zp = np.loadtxt((impresources.files(filter_data) / 'gaia_edr3_zeropt.dat'))
        transmission_curve[transmission_curve == 99.99] = 0.
        super().__init__(jnp.array([transmission_curve[0]*10., transmission_curve[5]]), name='Gaia RP', AB_zeropoint=zp[1, 4], Vega_zeropoint=zp[0, 4])
        
        
# Sartoretti et al. (2022)
class GaiaRVS(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) /'grvsfilter.csv'), skiprows=1, delimiter=',')
        super().__init__(jnp.array([transmission_curve[0]*10., transmission_curve[1]]), name='Gaia RVS', AB_zeropoint=21.317)

# Doi et al. (2010)

sdss = np.loadtxt((impresources.files(filter_data) / 'SDSS.dat'))

sdss_u = sdss[:, [0, 1]]
sdss_u[:, 1] *= sdss[:, -1]

sdss_g = sdss[:, [0, 2]]
sdss_g[:, 1] *= sdss[:, -1]

sdss_r = sdss[:, [0, 3]]
sdss_r[:, 1] *= sdss[:, -1]

sdss_i = sdss[:, [0, 4]]
sdss_i[:, 1] *= sdss[:, -1]

sdss_z = sdss[:, [0, 5]]
sdss_z[:, 1] *= sdss[:, -1]

class SDSSu(Filter):
    def __init__(self):
        super().__init__(sdss_u.T, name='SDSS u')


class SDSSg(Filter):
    def __init__(self):
        super().__init__(sdss_g.T, name='SDSS g')


class SDSSr(Filter):
    def __init__(self):
        super().__init__(sdss_r.T, name='SDSS r')


class SDSSi(Filter):
    def __init__(self):
        super().__init__(sdss_i.T, name='SDSS i')


class SDSSz(Filter):
    def __init__(self):
        super().__init__(sdss_z.T, name='SDSS z')


class TWOMASSJ(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / '2MASSJ.dat')).T
        transmission_curve[0] *= 1e4
        super().__init__(transmission_curve, name='2MASS J', non_photonic=True, Vega_zeropoint=-23.76248609213452 - 0.017)


class TWOMASSH(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / '2MASSH.dat')).T
        transmission_curve[0] *= 1e4
        super().__init__(transmission_curve, name='2MASS H', non_photonic=True, Vega_zeropoint=-24.84542522534151 + 0.016)


class TWOMASSK(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / '2MASSK.dat')).T
        transmission_curve[0] *= 1e4
        super().__init__(transmission_curve, name='2MASS Ks', non_photonic=True, Vega_zeropoint=-25.937629814008577 + 0.003)


# Morrissey et. al. (2004)
# https://cosmos.astro.caltech.edu/page/filterset


class GALEXFUV(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / 'GALEXFUV.dat')).T
        super().__init__(transmission_curve, name='GALEX FUV')


class GALEXNUV(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / 'GALEXNUV.dat')).T
        super().__init__(transmission_curve, name='GALEX NUV')


# https://github.com/lsst/throughputs/blob/main/baseline/README.md
# Just assume photonic with AB definition

class LSSTu(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / 'LSSTu.dat')).T
        super().__init__(transmission_curve, name='LSST u')


class LSSTg(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / 'LSSTg.dat')).T
        super().__init__(transmission_curve, name='LSST g')


class LSSTr(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / 'LSSTr.dat')).T
        super().__init__(transmission_curve, name='LSST r')


class LSSTi(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / 'LSSTi.dat')).T
        super().__init__(transmission_curve, name='LSST i')


class LSSTz(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / 'LSSTz.dat')).T
        super().__init__(transmission_curve, name='LSST z')


class LSSTy(Filter):
    def __init__(self):
        transmission_curve = np.loadtxt((impresources.files(filter_data) / 'LSSTy.dat')).T
        super().__init__(transmission_curve, name='LSST y')


# http://ipp.ifa.hawaii.edu/ps1.filters/
# Tonry et al. (2012)

panstarrs = np.loadtxt((impresources.files(filter_data) / 'panstarrs.dat'))
panstarrs[:, 0] *= 10.

panstarrs_open = panstarrs[:, [0, 1]]
panstarrs_g = panstarrs[:, [0, 2]]
panstarrs_r = panstarrs[:, [0, 3]]
panstarrs_i = panstarrs[:, [0, 4]]
panstarrs_z = panstarrs[:, [0, 5]]
panstarrs_y = panstarrs[:, [0, 6]]
panstarrs_w = panstarrs[:, [0, 7]]

class PANSTARRS_PS1_g(Filter):
    def __init__(self):
        super().__init__(panstarrs_g.T, name='PANSTARRS PS1 g', non_photonic=True)


class PANSTARRS_PS1_i(Filter):
    def __init__(self):
        super().__init__(panstarrs_i.T, name='PANSTARRS PS1 i', non_photonic=True)


class PANSTARRS_PS1_r(Filter):
    def __init__(self):
        super().__init__(panstarrs_r.T, name='PANSTARRS PS1 r', non_photonic=True)


class PANSTARRS_PS1_w(Filter):
    def __init__(self):
        super().__init__(panstarrs_w.T, name='PANSTARRS PS1 w', non_photonic=True)


class PANSTARRS_PS1_y(Filter):
    def __init__(self):
        super().__init__(panstarrs_y.T, name='PANSTARRS PS1 y', non_photonic=True)


class PANSTARRS_PS1_z(Filter):
    def __init__(self):
        super().__init__(panstarrs_z.T, name='PANSTARRS PS1 z', non_photonic=True)


class PANSTARRS_PS1_open(Filter):
    def __init__(self):
        super().__init__(panstarrs_open.T, name='PANSTARRS PS1 open', non_photonic=True)

# Source: Bessel (2011)

class Stromgrenu(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [3150., 3175., 3200., 3225., 3250., 3275., 3300., 3325., 3350., 3375., 3400., 3425., 3450., 3475., 3500., 3525., 3550., 3575., 3600., 3625., 3650., 3675., 3700., 3725., 3750., 3775., 3800., 3825., 3850.],
            [0.000, 0.004, 0.050, 0.122, 0.219, 0.341, 0.479, 0.604, 0.710, 0.809, 0.886, 0.939, 0.976, 1.000, 0.995, 0.981, 0.943, 0.880, 0.782, 0.659, 0.525, 0.370, 0.246, 0.151, 0.071, 0.030, 0.014, 0.000, 0.000]
        ])
        super().__init__(transmission_curve, name='Stromgren u', AB_zeropoint=-22.40788262039795+0.308)


class Stromgrenv(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [3750., 3775., 3800., 3825., 3850., 3875., 3900., 3925., 3950., 3975., 4000., 4025., 4050., 4075., 4100., 4125., 4150., 4175., 4200., 4225., 4250., 4275., 4300., 4325., 4350., 4375., 4400., 4425., 4450.],
            [0.000, 0.003, 0.006, 0.016, 0.029, 0.044, 0.060, 0.096, 0.157, 0.262, 0.404, 0.605, 0.810, 0.958, 1.000, 0.973, 0.882, 0.755, 0.571, 0.366, 0.224, 0.134, 0.079, 0.053, 0.039, 0.027, 0.014, 0.006, 0.000]
        ])
        super().__init__(transmission_curve, name='Stromgren v', AB_zeropoint=-22.40788262039795+0.327)


class Stromgrenb(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [4350., 4375., 4400., 4425., 4450., 4475., 4500., 4525., 4550., 4575., 4600., 4625., 4650., 4675., 4700., 4725., 4750., 4775., 4800., 4825., 4850., 4875., 4900., 4925., 4950., 4975., 5000., 5025., 5050.],
            [0.000, 0.010, 0.023, 0.039, 0.056, 0.086, 0.118, 0.188, 0.287, 0.457, 0.681, 0.896, 0.998, 1.000, 0.942, 0.783, 0.558, 0.342, 0.211, 0.130, 0.072, 0.045, 0.027, 0.021, 0.015, 0.011, 0.007, 0.003, 0.000]
        ])
        super().__init__(transmission_curve, name='Stromgren b', AB_zeropoint=-22.40788262039795+0.187)


class Stromgreny(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [5150., 5175., 5200., 5225., 5250., 5275., 5300., 5325., 5350., 5375., 5400., 5425., 5450., 5475., 5500., 5525., 5550., 5575., 5600., 5625., 5650., 5675., 5700., 5725., 5750., 5775., 5800., 5825., 5850.],
            [0.000, 0.022, 0.053, 0.082, 0.116, 0.194, 0.274, 0.393, 0.579, 0.782, 0.928, 0.985, 0.999, 1.000, 0.997, 0.938, 0.789, 0.574, 0.388, 0.232, 0.143, 0.090, 0.054, 0.031, 0.016, 0.010, 0.009, 0.004, 0.000]
        ])
        super().__init__(transmission_curve, name='Stromgren y', AB_zeropoint=-22.40788262039795+0.027)


class Bolometric(Filter):
    def __init__(self):
        transmission_curve = jnp.array([
            [1., 30000.],
            [1., 1.]
        ])
        super().__init__(transmission_curve, name='Bolometric')


FILTERS = [
    JohnsonCousinsU, JohnsonCousinsB, JohnsonCousinsV, JohnsonCousinsR, JohnsonCousinsI,
    TychoBT, TychoVT,
    GaiaG, GaiaRP, GaiaBP,
    SDSSu, SDSSg, SDSSr, SDSSi, SDSSz,
    TWOMASSJ, TWOMASSH, TWOMASSK,
    GALEXFUV, GALEXNUV,
    LSSTu, LSSTg, LSSTr, LSSTi, LSSTz, LSSTy,
    PANSTARRS_PS1_g, PANSTARRS_PS1_i, PANSTARRS_PS1_r, PANSTARRS_PS1_w, PANSTARRS_PS1_y, PANSTARRS_PS1_z, PANSTARRS_PS1_open,
    HipparcosHp,
    Stromgrenu, Stromgrenv, Stromgrenb, Stromgreny,
    Bolometric
]


FILTER_NAMES = [f.__name__ for f in FILTERS]



import jax
import jax.numpy as jnp
from jax.typing import ArrayLike


def sort_xy(points: ArrayLike) -> ArrayLike:
    """Sort points clockwise

    Args:
        points (ArrayLike): x, y coordinates (n, 2)

    Returns:
        ArrayLike: points with x, y coordinates sorted clockwise
    """
    x, y = points[:, 0], points[:, 1]
    x0 = jnp.mean(x)
    y0 = jnp.mean(y)

    r = jnp.sqrt((x-x0)**2 + (y-y0)**2)

    angles = jnp.where((y-y0) > 0, jnp.arccos((x-x0)/r), 2*jnp.pi-jnp.arccos((x-x0)/r))

    mask = jnp.argsort(angles)

    return points[mask[::-1], :] # clockwise order



def inside(p1: ArrayLike, p2: ArrayLike, q: ArrayLike) -> bool:
    """Check if the point is inside the polygon edge

    Args:
        p1 (ArrayLike): Polygon edge point 1
        p2 (ArrayLike): Polygon edge point 2
        q (ArrayLike): Point to be checked

    Returns:
        bool: Point is inside polygon
    """
    cross = (p2[0]-p1[0])*(q[1]-p1[1]) - (p2[1]-p1[1])*(q[0]-p1[0])
    return cross <= 0  # clockwise


def last_non_nan_arg(arr: ArrayLike) -> ArrayLike:
    """Return the last non-nan index of the array

    Args:
        arr (ArrayLike): Array

    Returns:
        ArrayLike: Last non-nan index of the array
    """    
    return jnp.max(jnp.argwhere(jnp.all(~jnp.isnan(arr), axis=1), size=arr.shape[0], fill_value=0)).astype(int)


def last_non_nan(arr: ArrayLike, axis=1) -> ArrayLike:
    """Return the last non-nan value of the array

    Args:
        arr (ArrayLike): Array

    Returns:
        ArrayLike: Last non-nan row of the array
    """    
    return arr[jnp.max(jnp.argwhere(jnp.all(~jnp.isnan(arr), axis=axis), size=arr.shape[0], fill_value=0))]


def append_to_last_nan(arr: ArrayLike, to_add: ArrayLike) -> ArrayLike:
    """Append the row to the first nan row

    Args:
        arr (ArrayLike): Array to add the row to
        to_add (ArrayLike): The row to be added

    Returns:
        ArrayLike: Array with an appended row
    """    
    next_ind = jnp.min(jnp.argwhere(jnp.all(jnp.isnan(arr), axis=1), size=arr.shape[0], fill_value=arr.shape[0]-1))
    return arr.at[next_ind].set(to_add.flatten())


def append_value_to_last_nan(arr: ArrayLike, to_add: ArrayLike) -> ArrayLike:
    """Append a single value to the first nan in array

    Args:
        arr (ArrayLike): Array to add the value to
        to_add (ArrayLike): The value to be appended

    Returns:
        ArrayLike: Array with an appended value
    """    
    next_ind = jnp.min(jnp.argwhere(jnp.isnan(arr), size=arr.shape[0], fill_value=arr.shape[0]-1))
    return arr.at[next_ind].set(to_add)


def append_value_to_last_nan_2d(arr: ArrayLike, to_add: ArrayLike) -> ArrayLike:
    """Append a single value to the first nan in array

    Args:
        arr (ArrayLike): Array to add the value to
        to_add (ArrayLike): The value to be appended

    Returns:
        ArrayLike: Array with an appended value
    """
    next_ind = jnp.min(jnp.argwhere(jnp.all(jnp.isnan(arr), axis=1), size=arr.shape[0], fill_value=arr.shape[0]-1))
    return arr.at[next_ind].set(to_add)


def repeat_last(arr: ArrayLike) -> ArrayLike:
    """Repeat the last non-nan row to fill the whole array

    Args:
        arr (ArrayLike):

    Returns:
        ArrayLike: Array with nan values filled with last non-nan values
    """    
    return jax.lax.cond(jnp.any(jnp.all(jnp.isnan(arr), axis=1)),
                        lambda: jax.lax.fori_loop(0, jnp.sum(jnp.all(jnp.isnan(arr), axis=1)),
                                                  lambda i, _arr: append_to_last_nan(_arr, last_non_nan(_arr)),
                                                  arr),
                        lambda: arr)


def wrap(arr: ArrayLike) -> ArrayLike:
    """Copy the first row and concatenate to the first array

    Args:
        arr (ArrayLike):

    Returns:
        ArrayLike: Array with the copy of the first row appended
    """    
    return jnp.concatenate([arr, arr[0].reshape((-1, arr.shape[1]))])


def polygon_area(x: ArrayLike, y: ArrayLike) -> ArrayLike:
    """Calculate a polygon's surface area from its x and y coordinates

    Args:
        x (ArrayLike): x coordinates of the polygon
        y (ArrayLike): y coordinates of the polygon

    Returns:
        ArrayLike: polygon's surface area
    """    
    return 0.5*jnp.abs(jnp.dot(x,jnp.roll(y,1))-jnp.dot(y,jnp.roll(x,1)))


polygon_areas = jax.jit(jax.vmap(polygon_area, in_axes=(0, 0)))

def cast_indexes(cast_vertices: ArrayLike) -> ArrayLike:
    """When casting 3D vectors to 2D, get the indices of non-reduced dimensions

    Args:
        cast_vertices (ArrayLike): vertices after 3D to 2D casting

    Returns:
        ArrayLike: non-reduced (non-zero) dimensions
    """
    zeroth_ind = jnp.argmax(jnp.all(jnp.isclose(cast_vertices, 0), axis=1))

    return jax.lax.cond(zeroth_ind == 0,
                        lambda: jnp.array([1, 2]),
                        lambda: jax.lax.cond(zeroth_ind == 1,
                                     lambda: jnp.array([0, 2]),
                                     lambda: jnp.array([0, 1])))


def get_cast_areas(face_vertices: ArrayLike) -> ArrayLike:
    """Get visible, 2D-casted polygon areas

    Args:
        face_vertices (ArrayLike): polygons' vertices after 3D to 2D casting

    Returns:
        ArrayLike: areas of 2D-casted polygons
    """
    return polygon_areas(face_vertices[:, :, 0],
                         face_vertices[:, :, 1])



import jax
import jax.numpy as jnp
from jax.typing import ArrayLike
from spice.geometry.utils import inside, last_non_nan, append_to_last_nan, repeat_last, sort_xy


@jax.jit
def x_y_first_line_vertical(p1, p2, q1, q2):
    x = p1[0]
    m2 = (q2[1]-q1[1])/(q2[0]-q1[0])
    b2 = q1[1]-m2*q1[0]
    return x, m2*x+b2

@jax.jit
def x_y_second_line_vertical(p1, p2, q1, q2):
    x = q1[0]
    m1 = (p2[1]-p1[1])/(p2[0]-p1[0])
    b1 = p1[1]-m1*p1[0]
    return x, m1*x+b1

@jax.jit
def x_y_no_vertical(p1, p2, q1, q2):
    m1 = (p2[1]-p1[1])/(p2[0]-p1[0])
    b1 = p1[1]-m1*p1[0]

    m2 = (q2[1]-q1[1])/(q2[0]-q1[0])
    b2 = q1[1]-m2*q1[0]

    x = (b2-b1)/(m1-m2)
    return x, m1*x+b1


@jax.jit
def compute_intersection(p1: ArrayLike, p2: ArrayLike, q1: ArrayLike, q2: ArrayLike) -> ArrayLike:
    """Calculate the intersection of two lines

    Args:
        p1 (ArrayLike): point 1 of line 1
        p2 (ArrayLike): point 2 of line 1
        q1 (ArrayLike): point 1 of line 2
        q2 (ArrayLike): point 2 of line 2

    Returns:
        ArrayLike: coordinates of the intersection of two lines
    """
    x, y = jax.lax.cond(jnp.isclose(p1[0]-p2[0], 0),
                        x_y_first_line_vertical,
                        lambda a, b, c, d: jax.lax.cond(
                            jnp.isclose(c[0]-d[0], 0),
                            x_y_second_line_vertical,
                            x_y_no_vertical,
                            a, b, c, d),
                        p1, p2, q1, q2)
    
    return jnp.array([[x, y]])

def s_only_edge_start_inside(final_polygon, c_edge_start, c_edge_end, s_edge_start, s_edge_end):
    intersection = compute_intersection(s_edge_start,
                                        s_edge_end,
                                        c_edge_start,
                                        c_edge_end)
    intersection = jax.lax.cond(jnp.any(jnp.isnan(intersection)),
                                lambda: jnp.nan*intersection,
                                lambda: intersection)
    return append_to_last_nan(final_polygon, intersection)

def s_only_edge_end_inside(final_polygon, c_edge_start, c_edge_end, s_edge_start, s_edge_end):
    intersection = compute_intersection(s_edge_start,
                                        s_edge_end,
                                        c_edge_start,
                                        c_edge_end)
    intersection = jax.lax.cond(jnp.any(jnp.isnan(intersection)),
                            lambda: jnp.nan*intersection,
                            lambda: intersection)
    polygon_with_intersection = append_to_last_nan(final_polygon, intersection)
    return append_to_last_nan(polygon_with_intersection, s_edge_end)

def s_both_start_and_end_inside(final_polygon, c_edge_start, c_edge_end, s_edge_start, s_edge_end):
    return append_to_last_nan(final_polygon, s_edge_end)


@jax.jit
def clip(subject_polygon: ArrayLike, clipping_polygon: ArrayLike) -> ArrayLike:
    """Calculate the intersection of the triangles subject_polygon and clipping_polygon

    Args:
        subject_polygon (ArrayLike): x and y coordinates (3, 2)
        clipping_polygon (ArrayLike): x and y coordinates (3, 2)

    Returns:
        ArrayLike: intersection between two triangles (6, 2)
    """ 
    subject_polygon = sort_xy(subject_polygon)
    clipping_polygon = sort_xy(clipping_polygon)
    final_polygon = repeat_last(jnp.concatenate([jnp.copy(subject_polygon), jnp.nan*jnp.ones((12-subject_polygon.shape[0], 2))], axis=0))

    def outer_iteration(i, final_polygon):
        next_polygon = jnp.copy(final_polygon)

        final_polygon = jnp.nan*jnp.ones((12, 2))

        #c_edge_start = clipping_polygon[((i-1)%(clipping_polygon.shape[0])).astype(int)]
        c_edge_start = jax.lax.cond(i==0,
                                    lambda: last_non_nan(clipping_polygon),
                                    lambda: clipping_polygon[i-1])
        c_edge_end = clipping_polygon[i]

        def inner_iteration(j, final_polygon):
            s_edge_start = jax.lax.cond(j==0,
                                        lambda: last_non_nan(next_polygon),
                                        lambda: next_polygon[j-1])
            s_edge_end = next_polygon[j]

            return jax.lax.cond(
                jnp.any(inside(c_edge_start, c_edge_end, s_edge_end)),
                lambda: jax.lax.cond(
                    jnp.any(inside(c_edge_start, c_edge_end, s_edge_start)),
                    lambda: s_both_start_and_end_inside(
                        final_polygon,
                        c_edge_start, c_edge_end,
                        s_edge_start, s_edge_end
                    ),
                    lambda: s_only_edge_end_inside(
                        final_polygon,
                        c_edge_start, c_edge_end,
                        s_edge_start, s_edge_end
                    ),
                ),
                lambda: jax.lax.cond(
                    jnp.any(inside(c_edge_start, c_edge_end, s_edge_start)),
                    lambda: s_only_edge_start_inside(
                        final_polygon,
                        c_edge_start, c_edge_end,
                        s_edge_start, s_edge_end
                    ),
                    lambda: final_polygon
                )
            )
        
        return jax.lax.fori_loop(0, next_polygon.shape[0], inner_iteration, final_polygon)
    return repeat_last(jax.lax.fori_loop(0, clipping_polygon.shape[0], outer_iteration, final_polygon))



import jax.numpy as jnp
import jax
from jax.typing import ArrayLike
from typing import Tuple

from jaxtyping import Array, Float


def vertex_to_polar(v: Float[Array, "3"]) -> Float[Array, "2"]:
    v += 1e-5
    r = jnp.sqrt(v[0] ** 2 + v[1] ** 2 + v[2] ** 2) + 1e-5
    return jnp.nan_to_num(
        jnp.array([
            jnp.arctan2(v[2], r),
            jnp.arctan2(v[1], v[0])
        ])
    )


def lat_to_theta(lat_deg: float) -> float:
    """
    Convert latitude in degrees to theta (inclination) in radians.
    
    Args:
        lat_deg (float): Latitude in degrees (-90 to 90)
        
    Returns:
        float: Theta (inclination) in radians (0 to pi)
    """
    return jnp.deg2rad(90 - lat_deg)  # Convert latitude to inclination angle (0 to pi)


def lon_to_phi(lon_deg: float) -> float:
    """
    Convert longitude in degrees to phi (azimuth) in radians.
    
    Args:
        lon_deg (float): Longitude in degrees (-180 to 180)
        
    Returns:
        float: Phi (azimuth) in radians (-pi to pi)
    """
    return jnp.deg2rad(lon_deg)  # Convert longitude to azimuth angle (-pi to pi)


def theta_to_lat(theta: float) -> float:
    """
    Convert theta (inclination) in radians to latitude in degrees.
    
    Args:
        theta (float): Inclination angle in radians (0 to pi)
        
    Returns:
        float: Latitude in degrees (-90 to 90)
    """
    return 90 - jnp.rad2deg(theta)  # Convert inclination to latitude


def phi_to_lon(phi: float) -> float:
    """
    Convert phi (azimuth) in radians to longitude in degrees.
    
    Args:
        phi (float): Azimuth angle in radians (-pi to pi)
        
    Returns:
        float: Longitude in degrees (-180 to 180)
    """
    return jnp.rad2deg(phi)  # Convert azimuth to longitude


@jax.jit
def mesh_polar_vertices(vertices: Float[Array, "n_vertices 3"]) -> Float[Array, "n_vertices 2"]:
    return (jax.vmap(vertex_to_polar, in_axes=0)(vertices))

@jax.jit
def spherical_harmonic(m, n, coordinates):
    polar_coordinates = mesh_polar_vertices(coordinates)
    m_array = (m * jnp.ones_like(polar_coordinates[:, 0])).astype(int)
    n_array = (n * jnp.ones_like(polar_coordinates[:, 1])).astype(int)
    return jax.scipy.special.sph_harm(m_array,
                                      n_array,
                                      polar_coordinates[:, 0],  # azimuthal angle (phi)
                                      polar_coordinates[:, 1],  # polar angle (theta)
                                      n_max=10).real


@jax.jit
def spherical_harmonic_with_tilt(m, n, coordinates, tilt_axis=jnp.array([0., 0., 1.]), tilt_angle=0.):
    # Normalize tilt axis
    tilt_axis = tilt_axis / jnp.linalg.norm(tilt_axis)
    
    r_matrix = evaluate_rotation_matrix(rotation_matrix(tilt_axis), jnp.deg2rad(tilt_angle))
    rotated_coords = jnp.matmul(coordinates, r_matrix)

    return spherical_harmonic(m, n, rotated_coords)


def evaluate_fourier_for_value(P: float, d: Float[Array, "1 n_orders"], phi: Float[Array, "1 n_orders"], timestamp: float) -> Float[Array, "1 n_orders"]:
    """
    Args:
        d0 (float): amplitude D_0
        P (float): period P
        d (Float[Array, "1 n_orders"]): amplitudes [1, n]
        phi (Float[Array, "1 n_orders"]): phases [1, n]
        timestamp (float): timestamps

    Returns:
        ArrayLike: values
    """
    n = jnp.arange(1, d.shape[0] + 1)
    return jnp.nan_to_num(jnp.sum(d * jnp.cos(2 * jnp.pi * n / P * timestamp - phi)))


def evaluate_fourier_prim_for_value(P: float, d: Float[Array, "1 n_orders"], phi: Float[Array, "1 n_orders"], timestamp: float) -> Float[Array, "1 n_orders"]:
    """
    Args:
        d0 (float): amplitude D_0
        P (float): period P
        d (Float[Array, "1 n_orders"]): amplitudes [1, n]
        phi (Float[Array, "1 n_orders"]): phases [1, n]
        timestamp (float): timestamps

    Returns:
        ArrayLike: values
    """
    n = jnp.arange(1, d.shape[0] + 1)
    return jnp.nan_to_num(jnp.sum(-2 * jnp.pi * d * n / P * jnp.sin(2 * jnp.pi * n / P * timestamp - phi)))


evaluate_many_fouriers_for_value = jax.jit(jax.vmap(evaluate_fourier_for_value, in_axes=(0, 0, 0, None)))
evaluate_many_fouriers_prim_for_value = jax.jit(jax.vmap(evaluate_fourier_prim_for_value, in_axes=(0, 0, 0, None)))


@jax.jit
def rotation_matrix(rotation_axis: Float[Array, "3"]) -> Float[Array, "3 3"]:
    a_norm = rotation_axis / jnp.linalg.norm(rotation_axis)
    a_hat = jnp.array([[0., -a_norm[2], a_norm[1]],
                       [a_norm[2], 0., -a_norm[0]],
                       [-a_norm[1], a_norm[0], 0.]])
    return a_hat


@jax.jit
def evaluate_rotation_matrix(rotation_matrix: Float[Array, "3 3"], theta: float) -> Float[Array, "3 3"]:
    return jnp.eye(3) + jnp.sin(theta) * rotation_matrix + (1 - jnp.cos(theta)) * jnp.matmul(rotation_matrix,
                                                                                             rotation_matrix)


@jax.jit
def rotation_matrix_prim(rotation_axis: Float[Array, "3"]) -> Float[Array, "3 3"]:
    a_norm = rotation_axis / jnp.linalg.norm(rotation_axis)
    a_hat = jnp.array([[0., -a_norm[2], a_norm[1]],
                       [a_norm[2], 0., -a_norm[0]],
                       [-a_norm[1], a_norm[0], 0.]])
    return a_hat


@jax.jit
def evaluate_rotation_matrix_prim(rotation_matrix_grad: Float[Array, "3 3"], theta: float) -> Float[Array, "3 3"]:
    return jnp.cos(theta) * rotation_matrix_grad + jnp.sin(theta) * jnp.matmul(rotation_matrix_grad,
                                                                               rotation_matrix_grad)


@jax.jit
def cast_to_los(vectors: Float[Array, "batch 3"], los_vector: Float[Array, "3"]) -> Float[Array, "batch"]:
    """Cast 3D vectors to the line-of-sight

    Args:
        vectors (Float[Array, "batch 3"]): Properties to be casted (n, 3)
        los_vector (Float[Array, "3"]): LOS vector (3,)

    Returns:
        ArrayLike: Casted vectors (n, 1)
    """
    return -1. * jnp.dot(vectors, los_vector)


@jax.jit
def cast_to_normal_plane(vectors: Float[Array, "batch 3"], los_vector: Float[Array, "3"]) -> Float[Array, "batch 2"]:
    """Cast 3D vectors to a 2D plane determined by a normal vector

    Args:
        vectors (Float[Array, "batch 3"]): Properties to be casted (n, 3)
        los_vector (Float[Array, "3"]): LOS vector (3,)

    Returns:
        ArrayLike: Casted vectors (n, 2)
    """
    """Cast 3D vectors to a 2D plane determined by the line-of-sight vector"""
    # Calculate the normal vector from the line-of-sight vector
    n = los_vector / jnp.linalg.norm(los_vector)
    
    # Create two orthogonal vectors in the plane perpendicular to the los_vector
    v1 = jnp.array([n[1], -n[0], 0])
    v1 = v1 / jnp.linalg.norm(v1)
    v2 = jnp.cross(n, v1)
    
    # Project the vectors onto the plane
    x = jnp.dot(vectors, v1)
    y = jnp.dot(vectors, v2)
    
    return jnp.column_stack((x, y))


@jax.jit
def cast_normalized_to_los(vectors: Float[Array, "batch 3"], los_vector: Float[Array, "3"]) -> Float[Array, "batch"]:
    """Cast 3D vectors to the line-of-sight

    Args:
        vectors (Float[Array, "batch 3"]): Properties to be casted (n, 3)
        los_vector (Float[Array, "3"]): LOS vector (3,)

    Returns:
        ArrayLike: Casted vectors (n, 1)
    """
    last_axis = len(vectors.shape) - 1
    return -1. * jnp.dot(vectors / (jnp.linalg.norm(vectors, axis=last_axis, keepdims=True) + 1e-10),
                         los_vector)


@jax.jit
def calculate_axis_radii(centers: Float[Array, "n_mesh_elements 3"], axis: Float[Array, "3"]) -> Float[Array, "n_mesh_elements"]:
    norm_axis = len(centers.shape) - 1
    return jnp.linalg.norm(jnp.cross(axis, -centers),
                           axis=norm_axis) / jnp.linalg.norm(axis)


### UTILITY FUNCTIONS FOR FOURIER SERIES

@jax.jit
def cos_as_fourier_coords(P: float, max_amplitude: float) -> Tuple[float, ArrayLike]:
    return P, jnp.array([[max_amplitude, 0.]])


@jax.jit
def sin_as_fourier_coords(P: float, max_amplitude: float) -> Tuple[float, ArrayLike]:
    return P, jnp.array([[max_amplitude, -jnp.pi / 2]])


@jax.jit
def sin2_as_fourier_coords(P: float, max_amplitude: float) -> Tuple[float, ArrayLike]:
    return 0.5 * P, jnp.array([[0.5 * max_amplitude, 0.]])


@jax.jit
def sinh_as_fourier_coords(P: float, max_amplitude: float, n: int) -> Tuple[float, ArrayLike]:
    return P, jnp.array(
        [[2 * jnp.sinh(jnp.pi) / jnp.pi * jnp.power(-1, n0 + 1) * n0 / (jnp.power(n0, 2) + 1) * max_amplitude, 0.] for
         n0 in range(1, n + 1)])



from typing import Optional, Union

import jax
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike
from spice.models.phoebe_model import PhoebeModel
from .mesh_model import MeshModel
from .utils import spherical_harmonic, spherical_harmonic_with_tilt

from jaxtyping import Array, Int, Float


def _is_arraylike(x):
    return hasattr(x, '__array__') or hasattr(x, '__array_interface__')


def __spherical_to_cartesian(theta: float, phi: float, radius: float) -> tuple[Array, Array, Array]:
    x = jnp.sin(theta) * jnp.cos(phi) * radius
    y = jnp.sin(theta) * jnp.sin(phi) * radius
    z = jnp.cos(theta) * radius
    return x, y, z


def __cartesian_distance(x1: float, y1: float, z1: float, x2: float, y2: float, z2: float) -> Array:
    return jnp.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2 + (z1 - z2) ** 2)


v_cartesian_distance = jax.jit(jax.vmap(__cartesian_distance, in_axes=(0, 0, 0, None, None, None)))


def sigmoid(x: float, smoothness: float) -> Array:
    # Apply the sigmoid function
    return 1 / (1 + jnp.exp(-smoothness * x))


@jax.jit
def generate_spherical_spot(d_centers: Float[Array, "n_mesh_elements 3"],
                            mesh_radius: float,
                            spot_center_theta: float,
                            spot_center_phi: float,
                            spot_radius: float,
                            parameter_diff: float,
                            smoothness: float = 1.0) -> Float[Array, "n_mesh_elements 3"]:
    """
    Generate a differential parameter array based on the location of points on a sphere relative to a specified
    spherical spot.

    This function calculates the angular distance of each point (defined by `thetas` and `phis`) from a specified
    spot center (`spot_center_theta`, `spot_center_phi`). It then determines the differential parameter value for
    each point based on whether it lies within the spherical spot, applying a smoothness factor to interpolate values
    at the spot's edge.

    Args:
        Float[Array, "n_mesh_elements 3"] (Array): Cartesian coordinates of the centers of the mesh elements
        mesh_radius (float): Radius of the mesh
        spot_center_theta (float): Theta value (inclination) of the spot center.
        spot_center_phi (float): Phi value (azimuth) of the spot center.
        spot_radius (float): Angular radius of the spot.
        parameter_diff (float): The differential parameter value to be applied within the spot.
        smoothness (float, optional): Factor controlling the smoothness of the spot's edge. Defaults to 1.0.

    Returns: ArrayLike: An array of differential parameter values for each point, interpolated based on spot
    membership and edge smoothness.
    """
    
    smoothness = jnp.clip(smoothness, 0.0, 1.0)

    spot_center_x, spot_center_y, spot_center_z = __spherical_to_cartesian(spot_center_theta, spot_center_phi,
                                                                           mesh_radius)

    # Calculate distance from the spot center
    distances = v_cartesian_distance(d_centers[:, 0], d_centers[:, 1], d_centers[:, 2],
                                     spot_center_x, spot_center_y, spot_center_z)

    # Calculate the difference from the spot edge
    distance_from_edge = jnp.deg2rad(spot_radius) * mesh_radius - distances

    # Apply smoothness to interpolate values at the spot's edge
    # Use a small fraction of mesh radius for scaling, but ensure it's not too small
    scale_factor = jnp.maximum(0.01 * mesh_radius, 1e-6)
    # Clip distance_from_edge to prevent numerical issues with very large negative values
    normalized_distance = jnp.clip(distance_from_edge / scale_factor, -100.0, 100.0)
    
    # Invert smoothness parameter so 0.0 is sharp (high sigmoid slope) and 1.0 is smooth (low sigmoid slope)
    # Map smoothness from [0.0, 1.0] to [10.0, 0.01] for sigmoid steepness
    adjusted_smoothness = (1.0 - smoothness) + 0.01 * smoothness
    differences = sigmoid(normalized_distance, adjusted_smoothness) * parameter_diff

    return differences


generate_spherical_spots = jax.jit(jax.vmap(generate_spherical_spot, in_axes=(None, None,
                                                                              0, 0, 0, 0, 0)))


@jax.jit
def _add_spherical_harmonic_spot(mesh: MeshModel,
                                 m: int, n: int,
                                 param_delta: float,
                                 param_index: int) -> MeshModel:
    spot_parameters = spherical_harmonic(m, n, mesh.centers) * param_delta
    return mesh._replace(parameters=mesh.parameters.at[:, param_index].set(
        mesh.parameters[:, param_index] + spot_parameters))
    
    
@jax.jit
def _add_spherical_harmonic_spot_with_tilt(mesh: MeshModel,
                                 m: int, n: int,
                                 param_delta: float,
                                 param_index: int,
                                 tilt_axis: ArrayLike = jnp.array([0., 0., 1.]),
                                 tilt_angle: float = 0.) -> MeshModel:
    spot_parameters = spherical_harmonic_with_tilt(m, n, mesh.centers, tilt_axis, tilt_angle) * param_delta
    return mesh._replace(parameters=mesh.parameters.at[:, param_index].set(
        mesh.parameters[:, param_index] + spot_parameters))


def add_spherical_harmonic_spot(mesh: MeshModel,
                                m_order: Union[Int, Float],
                                n_degree: Union[Int, Float],
                                param_delta: Float,
                                param_index: Float,
                                tilt_axis: Float[Array, "3"] = None,
                                tilt_angle: Float = None) -> MeshModel:
    """
    Add a spherical harmonic variation to a parameter of the mesh model, effectively creating a spot-like feature.

    This function applies a spherical harmonic function to modify a specific parameter of the mesh model,
    creating a pattern that can represent various surface features like spots or temperature variations.
    The spherical harmonic function Y_n^m(θ,φ) is used to create the variation pattern on the stellar surface.

    Args:
        mesh (MeshModel): The mesh model to which the spherical harmonic variation will be applied.
            Must not be a PhoebeModel as those are read-only.
        m_order (Union[Int, Float]): The order (m) of the spherical harmonic. Must be less than or equal 
            to n_degree. Determines the number of longitudinal nodes in the pattern.
        n_degree (Union[Int, Float]): The degree (n) of the spherical harmonic. Must be greater than or equal
            to m_order. Determines the total number of nodes in the pattern.
        param_delta (Float): The maximum amplitude of the parameter variation. Represents the difference 
            between the peak of the variation and the background value.
        param_index (Float): The index of the parameter in the mesh model's parameter array that will be 
            modified by this variation.
        tilt_axis (Float[Array, "3"], optional): The axis around which to tilt the spherical harmonic pattern.
            Defaults to None (no tilt).
        tilt_angle (Float, optional): The angle in degrees to tilt the spherical harmonic pattern around
            tilt_axis. Only used if tilt_axis is provided. Defaults to None (no tilt).

    Returns:
        MeshModel: A new mesh model with the specified parameter modified according to the spherical
            harmonic variation.

    Raises:
        ValueError: If mesh is a PhoebeModel, or if m_order > n_degree.
    """
    if isinstance(mesh, PhoebeModel):
        raise ValueError("PHOEBE models are read-only.")
    if isinstance(m_order, int) and isinstance(n_degree, int) and m_order > n_degree:
        raise ValueError("m must be lesser or equal to n.")
    elif _is_arraylike(m_order) and _is_arraylike(n_degree):
        m_order = int(m_order.item())
        n_degree = int(n_degree.item())
        if jnp.any(jnp.greater(m_order, n_degree)):
            raise ValueError("m must be lesser or equal to n.")
        
    if tilt_axis is not None:
        tilt_angle = 0.0 if tilt_angle is None else tilt_angle
        return _add_spherical_harmonic_spot_with_tilt(mesh, m_order, n_degree, param_delta, param_index, tilt_axis, tilt_angle)
    return _add_spherical_harmonic_spot(mesh, m_order, n_degree, param_delta, param_index)


@jax.jit
def _add_spot(mesh: MeshModel,
              spot_center_theta: float,
              spot_center_phi: float,
              spot_radius: float,
              parameter_diff: float,
              parameter_index: int,
              smoothness: float = 1.0) -> MeshModel:
    spot_parameters = generate_spherical_spot(mesh.d_centers, mesh.radius,
                                              spot_center_theta, spot_center_phi, spot_radius, parameter_diff,
                                              smoothness)
    return mesh._replace(parameters=mesh.parameters.at[:, parameter_index].set(
        mesh.parameters[:, parameter_index] + spot_parameters)
    )


def add_spot(mesh: MeshModel,
             spot_center_theta: float,
             spot_center_phi: float,
             spot_radius: float,
             parameter_delta: float,
             parameter_index: int,
             smoothness: float = 0.0) -> MeshModel:
    """
    Add a spot to a mesh model based on spherical coordinates and smoothness parameters.

    This function applies a modification to the mesh model's parameters to simulate the presence of a spot. The spot
    is defined by its center (in spherical coordinates), its radius, and a differential parameter that quantifies the
    change induced by the spot. The smoothness parameter allows for a gradual transition at the spot's edges.

    Args:
        mesh (MeshModel): The mesh model to which the spot will be added. Must not be a PHOEBE model, as they are read-only.
        spot_center_theta (float): The theta (inclination) coordinate of the spot's center, in radians.
        spot_center_phi (float): The phi (azimuthal) coordinate of the spot's center, in radians.
        spot_radius (float): The angular radius of the spot, in degrees.
        parameter_delta (float): The difference in the parameter value to be applied within the spot.
        parameter_index (int): The index of the parameter in the mesh model that will be modified by the spot.
        smoothness (float, optional): A parameter controlling the smoothness of the spot's edge. Higher values result in smoother transitions. Defaults to 0.0.

    Returns:
        MeshModel: The modified mesh model with the spot applied.

    Raises:
        ValueError: If the input mesh model is a PHOEBE model, as modifications are not supported.
    """
    if isinstance(mesh, PhoebeModel):
        raise ValueError("PHOEBE models are read-only.")
    else:
        return _add_spot(mesh, spot_center_theta, spot_center_phi, spot_radius,
                         parameter_delta, parameter_index, smoothness)


def add_spots(mesh: MeshModel,
              spot_center_thetas: Float[Array, "n_spots"],
              spot_center_phis: Float[Array, "n_spots"],
              spot_radii: Float[Array, "n_spots"],
              parameter_deltas: Float[Array, "n_spots"],
              parameter_indices: Int[Array, "n_spots"],
              smoothness: Float[Array, "n_spots"] = None) -> MeshModel:
    """
    Add multiple spots to a mesh model based on spherical coordinates and smoothness parameters.

    This function applies modifications to the mesh model's parameters to simulate the presence of multiple spots.
    Each spot is defined by its center (in spherical coordinates), its radius, and a differential parameter that
    quantifies the change induced by the spot. The smoothness parameter allows for a gradual transition at the spot's edges.

    Args:
        mesh (MeshModel): The mesh model to which the spots will be added. Must not be a PHOEBE model, as they are read-only.
        spot_center_thetas (Float[Array, "n_spots"]): An array of theta (inclination) coordinates of the spots' centers, in radians.
        spot_center_phis (Float[Array, "n_spots"]): An array of phi (azimuthal) coordinates of the spots' centers, in radians. 
        spot_radii (Float[Array, "n_spots"]): An array of angular radii of the spots, in degrees.
        parameter_deltas (Float[Array, "n_spots"]): An array of differences in the parameter values to be applied within the spots.
        parameter_indices (Int[Array, "n_spots"]): An array of indices of the parameters in the mesh model that will be modified by the spots.
        smoothness (Float[Array, "n_spots"], optional): An array of factors controlling the smoothness of the spots' edges. Higher values result in smoother transitions. Defaults to None.

    Returns:
        MeshModel: The modified mesh model with the spots applied.

    Raises:
        ValueError: If the input mesh model is a PHOEBE model, as modifications are not supported.
    """
    if isinstance(mesh, PhoebeModel):
        raise ValueError("PHOEBE models are read-only.")
    else:
        if parameter_indices is None:
            parameter_indices = jnp.ones_like(parameter_deltas)
        
        def scan_fn(carry, x):
            mesh, theta, phi, radius, delta, index, smooth = carry, *x
            return _add_spot(mesh, theta, phi, radius, delta, index, smooth), None

        initial_carry = mesh
        xs = (spot_center_thetas, spot_center_phis, spot_radii,
              parameter_deltas, parameter_indices, smoothness)
        
        final_mesh, _ = jax.lax.scan(scan_fn, initial_carry, xs)
        return final_mesh


@jax.jit
def _add_spherical_harmonic_spots(mesh: MeshModel,
                                  m: ArrayLike, n: ArrayLike,
                                  param_deltas: ArrayLike,
                                  param_indices: ArrayLike) -> MeshModel:
    def scan(carry, params):
        return _add_spherical_harmonic_spot(
            carry, m=params[0].astype(int), n=params[1].astype(int),
            param_delta=params[2], param_index=params[3].astype(int)
        ), params

    updated_mesh, _ = jax.lax.scan(scan, mesh, jnp.vstack([m, n, param_deltas, param_indices]).T)
    return updated_mesh


@jax.jit
def _add_spherical_harmonic_spots_with_tilt(mesh: MeshModel,
                                  m: ArrayLike, n: ArrayLike,
                                  param_deltas: ArrayLike,
                                  param_indices: ArrayLike,
                                  tilt_axes: ArrayLike,
                                  tilt_angles: ArrayLike) -> MeshModel:
    def scan(carry, params):
        # Extract tilt axis from params[4:7] since it's 3 components
        tilt_axis = params[4:7]
        return _add_spherical_harmonic_spot_with_tilt(
            carry, m=params[0].astype(int), n=params[1].astype(int),
            param_delta=params[2], param_index=params[3].astype(int),
            tilt_axis=tilt_axis, tilt_angle=params[7]
        ), params

    # Reshape tilt_axes to flatten the last dimension
    tilt_axes_flat = tilt_axes.reshape(-1)
    
    # Stack parameters, handling the flattened tilt axes
    params = jnp.vstack([
        m, n, param_deltas, param_indices,
        tilt_axes_flat[0::3],  # x components
        tilt_axes_flat[1::3],  # y components 
        tilt_axes_flat[2::3],  # z components
        tilt_angles
    ]).T

    updated_mesh, _ = jax.lax.scan(scan, mesh, params)
    return updated_mesh


def add_spherical_harmonic_spots(mesh: MeshModel,
                                 m_orders: Float[Array, "n_orders"],
                                 n_degrees: Float[Array, "n_orders"],
                                 param_deltas: Float[Array, "n_orders"],
                                 param_indices: Float[Array, "n_orders"],
                                 tilt_axes: Optional[Float[Array, "n_orders 3"]] = None,
                                 tilt_angles: Optional[Float[Array, "n_orders"]] = None) -> MeshModel:
    """
    Add multiple spherical harmonic spots to a mesh model.

    This function iteratively applies spherical harmonic modifications to a mesh model based on the provided parameters.
    Each spot is defined by spherical harmonic indices (m_order, n_degree), a parameter delta indicating the modification strength,
    and a parameter index specifying which parameter of the mesh model is modified. The function checks if the mesh model
    is not a PHOEBE model, as they are read-only, and raises an error if it is.

    Args:
        mesh (MeshModel): The mesh model to which the spots will be added.
        m_orders (Float[Array, "n_orders"]): An array of m indices (orders) for the spherical harmonics.
        n_degrees (Float[Array, "n_orders"]): An array of n indices (degrees) for the spherical harmonics.
        param_deltas (Float[Array, "n_orders"]): An array of deltas specifying the strength of the modification for each spot.
        param_indices (Float[Array, "n_orders"]): An array of parameter indices specifying which parameter of the mesh is modified by each spot.
        tilt_axes (Optional[Float[Array, "n_orders 3"]], optional): An array of tilt axes for each spot. Shape should be (n_orders, 3). Defaults to None (no tilt).
        tilt_angles (Optional[Float[Array, "n_orders"]], optional): An array of tilt angles in degrees for each spot. Defaults to None (no tilt).

    Returns:
        MeshModel: The modified mesh model with all spherical harmonic spots applied.

    Raises:
        ValueError: If the input mesh model is a PHOEBE model, as modifications are not supported.
        ValueError: If tilt_axes is provided without tilt_angles or vice versa.
    """
    if isinstance(mesh, PhoebeModel):
        raise ValueError("PHOEBE models are read-only.")
    if (tilt_axes is not None) and (tilt_angles is not None):
        return _add_spherical_harmonic_spots_with_tilt(mesh, m_orders, n_degrees, param_deltas, param_indices, tilt_axes, tilt_angles)
    else:
        return _add_spherical_harmonic_spots(mesh, m_orders, n_degrees, param_deltas, param_indices)



import jax
import jax.numpy as jnp
from jax import lax
from jax.typing import ArrayLike
from jax.scipy.integrate import trapezoid
from typing import Callable
from spice.models import MeshModel
import math
from functools import partial
from spice.spectrum.utils import ERG_S_TO_W
from spice.spectrum.filter import Filter

from jaxtyping import Array, Float

DEFAULT_CHUNK_SIZE: int = 1024
C: float = 299792.458  # km/s
SOL_RAD_CM = 69570000000.0

apply_vrad = lambda x, vrad: x * (vrad / C + 1)
apply_vrad_log = lambda x, vrad: x + jnp.log10(vrad / C + 1)
v_apply_vrad = jax.jit(jax.vmap(apply_vrad, in_axes=(None, 0)))

# n_wavelengths, n_vertices
v_apply_vrad_log = jax.jit(jax.vmap(apply_vrad_log, in_axes=(None, 0), out_axes=0))


@partial(jax.jit, static_argnums=(0, 6))
def __spectrum_flash_sum(intensity_fn,
                         log_wavelengths,
                         areas,
                         mus,
                         vrads,
                         parameters,
                         chunk_size: int,
                         disable_doppler_shift: bool = False):
    '''
    Each surface element has a vector of parameters (mu, LOS velocity, etc)
    Some of these parameters are the flux model's input
    '''

    # Just the 1D case for now
    n_areas = areas.shape[0]
    n_parameters = parameters.shape[-1]

    v_intensity = jax.vmap(intensity_fn, in_axes=(0, 0, 0))
    
    n = math.ceil(n_areas / chunk_size)

    @partial(jax.checkpoint, prevent_cse=False)
    def chunk_scanner(carries, x):
        chunk_idx, atmo_sum = carries

        k_chunk_sizes = min(chunk_size, n_areas)

        # (CHUNK_SIZE, 1)
        a_chunk = lax.dynamic_slice(areas,
                                    (chunk_idx,),
                                    (k_chunk_sizes,))

        # (CHUNK_SIZE, 1)
        m_chunk = lax.dynamic_slice(mus,
                                    (chunk_idx,),
                                    (k_chunk_sizes,))

        # (CHUNK_SIZE, 1)
        vrad_chunk = lax.dynamic_slice(vrads,
                                       (chunk_idx,),
                                       (k_chunk_sizes,))
        # (CHUNK_SIZE, n_parameters)
        p_chunk = lax.dynamic_slice(parameters,
                                    (chunk_idx, 0),
                                    (k_chunk_sizes, n_parameters))

        # Shape: (CHUNK_SIZE, log_wavelengths)
        shifted_log_wavelengths = jax.lax.cond(disable_doppler_shift,
                                               lambda lv, _: jnp.repeat(lv[jnp.newaxis, :], chunk_size, axis=0),
                                               v_apply_vrad_log,
                                               log_wavelengths, vrad_chunk)

        # atmosphere_mul is the spectrum simulated for the corresponding wavelengths and optionally given parameters of mu, logg, and T.
        # It is then multiplied by the observed area to scale the contributions of spectra chunks

        # Shape: (n_vertices, 2, n_wavelengths)
        # Areas should be rescaled by mus
        # 2 corresponds to the two components: continuum and full spectrum with lines
        # n_wavelengths, n_verices, 2 (continuum+spectrum), 1

        # shifted_log_wavelengths (CHUNK_SIZE, n_wavelengths)
        # m_chunk (CHUNK_SIZE)
        # p_chunk (CHUNK_SIZE, n_parameters)
        v_in = v_intensity(shifted_log_wavelengths,  # (n,)
                           m_chunk[:, jnp.newaxis],
                           p_chunk)
        atmosphere_mul = jnp.multiply(
            (a_chunk)[:, jnp.newaxis, jnp.newaxis],
            v_in)

        new_atmo_sum = atmo_sum + jnp.sum(atmosphere_mul, axis=0)

        return (chunk_idx + k_chunk_sizes, new_atmo_sum), chunk_idx+k_chunk_sizes

    # Return (2, n_vertices) for continuum and spectrum with lines
    (_, out), _ = lax.scan(
        chunk_scanner,
        init=(0, jnp.zeros((log_wavelengths.shape[-1], 2))),
        xs=None,
        length=n)
    return out


@partial(jax.jit, static_argnums=(0, 6, 7))
def __spectrum_flash_sum_with_padding(intensity_fn,
                         log_wavelengths,
                         areas,
                         mus,
                         vrads,
                         parameters,
                         chunk_size: int,
                         wavelengths_chunk_size: int,
                         disable_doppler_shift: bool = False):
    n_padding = wavelengths_chunk_size - (log_wavelengths.shape[0] % wavelengths_chunk_size)
    log_wavelengths_padded = jnp.pad(log_wavelengths, (0, n_padding), mode='constant', constant_values=0)
    # Reshape padded wavelengths into chunks
    wavelength_chunks = log_wavelengths_padded.reshape(-1, wavelengths_chunk_size).T
    
    def scan_fn(carry, chunk):
        result = __spectrum_flash_sum(intensity_fn,
                                    chunk,
                                    areas,
                                    mus,
                                    vrads,
                                    parameters,
                                    chunk_size,
                                    disable_doppler_shift)
        return carry, result

    _, results = lax.scan(scan_fn,
                         None,
                         wavelength_chunks)
    return results.reshape(-1, 2, order='F')


@partial(jax.jit, static_argnums=(1,))
def _adjust_dim(x: ArrayLike, chunk_size: int) -> ArrayLike:
    return jnp.concatenate([x, jnp.zeros((chunk_size - x.shape[0] % chunk_size, *x.shape[1:]))], axis=0)


# TODO: think: change to simulate_obseved_monochromatic_flux?
@partial(jax.jit, static_argnums=(0, 4, 5))
def simulate_observed_flux(intensity_fn: Callable[[Float[Array, "n_wavelengths"], float, Float[Array, "n_mesh_elements n_parameters"]], Float[Array, "n_wavelengths 2"]],
                           m: MeshModel,
                           log_wavelengths: Float[Array, "n_wavelengths"],
                           distance: float = 10.0,
                           chunk_size: int = DEFAULT_CHUNK_SIZE,
                           wavelengths_chunk_size: int = DEFAULT_CHUNK_SIZE,
                           disable_doppler_shift: bool = False) -> Float[Array, "n_wavelengths 2"]:
    """Simulate the observed flux from a mesh model.

    This function calculates the observed flux from a mesh model by combining the intensity function,
    mesh geometry, and radiative transfer effects. It accounts for visible areas, viewing angles (mu),
    line-of-sight velocities, and distance effects.

    Args:
        intensity_fn (Callable): Function that computes intensity given wavelengths, mu and parameters
        m (MeshModel): The mesh model containing geometry and physical parameters
        log_wavelengths (Float[Array, "n_wavelengths"]): Log of wavelength points to evaluate
        distance (float, optional): Distance to object in parsecs. Defaults to 10.0.
        chunk_size (int, optional): Size of chunks for parallel processing. Defaults to 1024.
        wavelengths_chunk_size (int, optional): Chunk size for wavelength array. Defaults to 1024.
        disable_doppler_shift (bool, optional): Whether to disable Doppler shift calculations. Defaults to False.

    Returns:
        Float[Array, "n_wavelengths 2"]: Array containing the computed flux at each wavelength point.
        The second dimension contains [flux, flux_error].
        Units are erg/s/cm^2/Å.
    """
    
    return jnp.nan_to_num(__spectrum_flash_sum_with_padding(intensity_fn,
                                               log_wavelengths,
                                               _adjust_dim(m.visible_cast_areas, chunk_size),
                                               _adjust_dim(jnp.where(m.mus > 0, m.mus, 0.), chunk_size),
                                               _adjust_dim(m.los_velocities, chunk_size),
                                               _adjust_dim(m.parameters, chunk_size),
                                               chunk_size,
                                               wavelengths_chunk_size,
                                               disable_doppler_shift) * jnp.power(m.radius,
                                                                                  2) * 5.08326693599739e-16 / (
                                      distance ** 2))[:len(log_wavelengths), :]


@partial(jax.jit, static_argnums=(0, 5))
def __flux_flash_sum(flux_fn,
                     log_wavelengths,
                     areas,
                     vrads,
                     parameters,
                     chunk_size: int,
                     disable_doppler_shift: bool = False):

    # Just the 1D case for now
    n_areas = areas.shape[0]
    n_parameters = parameters.shape[-1]

    v_flux = jax.vmap(flux_fn, in_axes=(0, 0))

    @partial(jax.checkpoint, prevent_cse=False)
    def chunk_scanner(carries, _):
        chunk_idx, atmo_sum = carries

        k_chunk_sizes = min(chunk_size, n_areas)

        # (CHUNK_SIZE, 1)
        a_chunk = lax.dynamic_slice(areas,
                                    (chunk_idx,),
                                    (k_chunk_sizes,))

        # (CHUNK_SIZE, 1)
        vrad_chunk = lax.dynamic_slice(vrads,
                                       (chunk_idx,),
                                       (k_chunk_sizes,))
        # (CHUNK_SIZE, n_parameters)
        p_chunk = lax.dynamic_slice(parameters,
                                    (chunk_idx, 0),
                                    (k_chunk_sizes, n_parameters))

        # Shape: (CHUNK_SIZE, log_wavelengths)
        shifted_log_wavelengths = jax.lax.cond(disable_doppler_shift,
                                               lambda lv, _: jnp.repeat(lv[jnp.newaxis, :], chunk_size, axis=0),
                                               v_apply_vrad_log,
                                               log_wavelengths, vrad_chunk)

        # atmosphere_mul is the spectrum simulated for the corresponding wavelengths and optionally given parameters of mu, logg, and T.
        # It is then multiplied by the observed area to scale the contributions of spectra chunks

        # Shape: (n_vertices, 2, n_wavelengths)
        # Areas should be rescaled by mus
        # 2 corresponds to the two components: continuum and full spectrum with lines
        # n_wavelengths, n_verices, 2 (continuum+spectrum), 1

        # shifted_log_wavelengths (CHUNK_SIZE, n_wavelengths)
        # m_chunk (CHUNK_SIZE)
        # p_chunk (CHUNK_SIZE, n_parameters)
        v_in = v_flux(shifted_log_wavelengths,  # (n,)
                      p_chunk)

        atmosphere_mul = jnp.multiply(
            a_chunk[:, jnp.newaxis, jnp.newaxis],  # Czemy nie 2D? Broadcastowanie?
            v_in)

        new_atmo_sum = atmo_sum + jnp.sum(atmosphere_mul, axis=0)

        return (chunk_idx + k_chunk_sizes, new_atmo_sum), None

    # Return (2, n_vertices) for continuum and spectrum with lines
    (_, out), _ = lax.scan(
        chunk_scanner,
        init=(0, jnp.zeros((log_wavelengths.shape[-1], 2))),
        xs=None,
        length=math.ceil(n_areas / chunk_size))
    return out


@partial(jax.jit, static_argnums=(0, 5, 6))
def __flux_flash_sum_with_padding(intensity_fn,
                         log_wavelengths,
                         areas,
                         vrads,
                         parameters,
                         chunk_size: int,
                         wavelengths_chunk_size: int,
                         disable_doppler_shift: bool = False):
    n_padding = wavelengths_chunk_size - (log_wavelengths.shape[0] % wavelengths_chunk_size)
    log_wavelengths_padded = jnp.pad(log_wavelengths, (0, n_padding), mode='constant', constant_values=0)
    # Reshape padded wavelengths into chunks
    wavelength_chunks = log_wavelengths_padded.reshape(-1, wavelengths_chunk_size).T
    
    def scan_fn(carry, chunk):
        result = __flux_flash_sum(intensity_fn,
                                    chunk,
                                    areas,
                                    vrads,
                                    parameters,
                                    chunk_size,
                                    disable_doppler_shift)
        return carry, result

    _, results = lax.scan(scan_fn,
                         None,
                         wavelength_chunks)
    return results.reshape(-1, 2, order='F')


@partial(jax.jit, static_argnums=(0, 3, 4))
def simulate_monochromatic_luminosity(flux_fn: Callable[[Float[Array, "n_wavelengths"], Float[Array, "n_mesh_elements n_parameters"]], Float[Array, "n_wavelengths 2"]],
                                      m: MeshModel,
                                      log_wavelengths: Float[Array, "n_wavelengths"],
                                      chunk_size: int = DEFAULT_CHUNK_SIZE,
                                      wavelengths_chunk_size: int = DEFAULT_CHUNK_SIZE,
                                      disable_doppler_shift: bool = False) -> Float[Array, "n_wavelengths 2"]:
    """Simulate the monochromatic luminosity from a mesh model.

    This function calculates the monochromatic luminosity from a mesh model by combining the flux function,
    mesh geometry, and radiative transfer effects. It accounts for visible areas, line-of-sight velocities,
    and radius effects.

    Args:
        flux_fn (Callable[[Float[Array, "n_wavelengths"], Float[Array, "n_mesh_elements n_parameters"]], Float[Array, "n_wavelengths 2"]]): 
            Function that computes flux given wavelengths and parameters
        m (MeshModel): The mesh model containing geometry and physical parameters
        log_wavelengths (Float[Array, "n_wavelengths"]): Log of wavelength points to evaluate
        chunk_size (int, optional): Size of chunks for parallel processing. Defaults to 1024.
        wavelengths_chunk_size (int, optional): Chunk size for wavelength array. Defaults to 1024.
        disable_doppler_shift (bool, optional): Whether to disable Doppler shift calculations. Defaults to False.

    Returns:
        Float[Array, "n_wavelengths 2"]: Array containing the computed luminosity at each wavelength point.
        The second dimension contains [luminosity, luminosity_error].
        Units are erg/s/Å.
    """
    return jnp.nan_to_num(__flux_flash_sum_with_padding(flux_fn,
                                           log_wavelengths,
                                           _adjust_dim(m.areas, chunk_size),
                                           _adjust_dim(m.los_velocities, chunk_size),
                                           _adjust_dim(m.parameters, chunk_size),
                                           chunk_size,
                                           wavelengths_chunk_size,
                                           disable_doppler_shift) * jnp.power(m.radius, 2) * 4.8399849e+21)[:len(log_wavelengths), :]


@partial(jax.jit, static_argnums=(0, 3, 4))
def luminosity(flux_fn: Callable[[Float[Array, "n_wavelengths"], Float[Array, "n_mesh_elements n_parameters"]], Float[Array, "n_wavelengths 2"]],
               model: MeshModel,
               wavelengths: Float[Array, "n_wavelengths"],
               chunk_size: int = DEFAULT_CHUNK_SIZE,
               wavelengths_chunk_size: int = DEFAULT_CHUNK_SIZE) -> float:
    """Calculate the bolometric luminosity of the model.

    This function computes the total bolometric luminosity by integrating the monochromatic luminosity
    over all wavelengths. It uses the trapezoidal rule for numerical integration.

    Args:
        flux_fn (Callable[[Float[Array, "n_wavelengths"], Float[Array, "n_mesh_elements n_parameters"]], Float[Array, "n_wavelengths 2"]]):
            Function that computes flux given wavelengths and parameters
        model (MeshModel): The mesh model containing geometry and physical parameters
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points to evaluate [Angstrom]
        chunk_size (int, optional): Size of chunks for parallel processing. Defaults to 1024.
        wavelengths_chunk_size (int, optional): Chunk size for wavelength array. Defaults to 1024.

    Returns:
        float: Total bolometric luminosity [erg/s]
    """
    flux = simulate_monochromatic_luminosity(flux_fn, model, jnp.log10(wavelengths), chunk_size, wavelengths_chunk_size)
    return trapezoid(y=flux[:, 0], x=wavelengths * 1e-8)


@jax.jit
def filter_responses(wavelengths: ArrayLike, sample_wavelengths: ArrayLike, sample_responses: ArrayLike) -> ArrayLike:
    return jnp.interp(wavelengths, sample_wavelengths, sample_responses)


# TODO: rename to observed passband luminosity? how to name it so that it's consistent???
# Let's rename it to obseved_passband_flux
# later let's have AB_observed_passband_magnitude

# If we set the transmission curve to be 1.0 for all wavelengths, we get the bolometric flux (NOT MONochormatic anymore)

@partial(jax.jit, static_argnums=(0,))
def __AB_passband_luminosity_photonic(filter: Filter,
                                      wavelengths: Float[Array, "n_wavelengths"],
                                      observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the AB magnitude in a given filter passband.

    This function computes the AB magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the AB magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^2/Å]

    Returns:
        float: AB magnitude in the filter passband [mag]
    """
    transmission_responses = filter.filter_responses_for_wavelengths(wavelengths)
    
    return -2.5 * jnp.log10(
            trapezoid(x=wavelengths * 1e-8, y=wavelengths * 1e-8 * observed_flux * transmission_responses) /
            (trapezoid(x=wavelengths * 1e-8, y=transmission_responses / (wavelengths * 1e-8)))
        ) + filter.AB_zeropoint
    
    
@partial(jax.jit, static_argnums=(0,))
def __AB_passband_luminosity_photonic_gaia(filter: Filter,
                                           wavelengths: Float[Array, "n_wavelengths"],
                                           observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the AB magnitude in a given filter passband.

    This function computes the AB magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the AB magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^2/Å]

    Returns:
        float: AB magnitude in the filter passband [mag]
    """
    transmission_responses = filter.filter_responses_for_wavelengths(wavelengths)
    # Montegriffo (2023)
    # P_A/(10^9 h*c) * int(lambda * f_lambda * T_zeta dlambda)
    
    nm_lambda = wavelengths/10.
    watt_flux = 1e-10*observed_flux

    return -2.5 * jnp.log10(3663830098681164.5* 
            trapezoid(x=nm_lambda, y=nm_lambda * watt_flux * transmission_responses)
    ) + filter.AB_zeropoint

@partial(jax.jit, static_argnums=(0,))
def __AB_passband_luminosity_non_photonic(filter: Filter,
                                          wavelengths: Float[Array, "n_wavelengths"],
                                          observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the AB magnitude in a given filter passband.

    This function computes the AB magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the AB magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^2/Å]

    Returns:
        float: AB magnitude in the filter passband [mag]
    """
    transmission_responses = filter.filter_responses_for_wavelengths(wavelengths)

    return -2.5 * jnp.log10(
            trapezoid(x=wavelengths, y=observed_flux / 1e8 * transmission_responses) /
            trapezoid(x=wavelengths, y=transmission_responses)
    ) + filter.AB_zeropoint
    
    
@partial(jax.jit, static_argnums=(0,))
def __AB_passband_luminosity_non_photonic_panstarrs_ps1(filter: Filter,
                                                        wavelengths: Float[Array, "n_wavelengths"],
                                                        observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the AB magnitude in a given filter passband for PanSTARRS PS1.

    This function computes the AB magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the AB magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^2/Å]

    Returns:
        float: AB magnitude in the filter passband [mag]
    """
    transmission_responses = filter.filter_responses_for_wavelengths(wavelengths)

    return -2.5 * jnp.log10(
            trapezoid(x=wavelengths, y=observed_flux * transmission_responses) /
            trapezoid(x=wavelengths, y=transmission_responses / ((wavelengths*1e-8)**2))
    ) + filter.AB_zeropoint

def AB_passband_luminosity(filter: Filter,
                           wavelengths: Float[Array, "n_wavelengths"],
                           observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the AB magnitude in a given filter passband.

    This function computes the AB magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the AB magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^2/Å]

    Returns:
        float: AB magnitude in the filter passband [mag]
    """
    if 'gaia' in filter.name.lower():
        return __AB_passband_luminosity_photonic_gaia(filter, wavelengths, observed_flux)
    elif 'panstarrs' in filter.name.lower():
        return __AB_passband_luminosity_non_photonic_panstarrs_ps1(filter, wavelengths, observed_flux)
    elif filter.non_photonic:
        return __AB_passband_luminosity_non_photonic(filter, wavelengths, observed_flux)
    else:
        return __AB_passband_luminosity_photonic(filter, wavelengths, observed_flux)


@partial(jax.jit, static_argnums=(0,))
def __ST_passband_luminosity_non_photonic(filter: Filter,
                                          wavelengths: Float[Array, "n_wavelengths"],
                                          observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the ST magnitude in a given filter passband.

    This function computes the ST magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the ST magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^2/Å]

    Returns:
        float: ST magnitude in the filter passband [mag]
    """
    transmission_responses = filter.filter_responses_for_wavelengths(wavelengths)
    
    return -2.5 * jnp.log10(
            trapezoid(x=wavelengths, y=observed_flux / 1e8 * transmission_responses) /
            trapezoid(x=wavelengths, y=transmission_responses)
        ) + filter.ST_zeropoint
    
    
@partial(jax.jit, static_argnums=(0,))
def __ST_passband_luminosity_non_photonic_panstarrs_ps1(filter: Filter,
                                                        wavelengths: Float[Array, "n_wavelengths"],
                                                        observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the ST magnitude in a given filter passband for PanSTARRS PS1.

    This function computes the ST magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the AB magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^2/Å]

    Returns:
        float: ST magnitude in the filter passband [mag]
    """
    transmission_responses = filter.filter_responses_for_wavelengths(wavelengths)

    return -2.5 * jnp.log10(
            trapezoid(x=wavelengths, y=observed_flux * transmission_responses) /
            trapezoid(x=wavelengths, y=transmission_responses / ((wavelengths*1e-8)**2))
    ) + filter.ST_zeropoint

    
@partial(jax.jit, static_argnums=(0,))
def __ST_passband_luminosity_photonic(filter: Filter,
                                      wavelengths: Float[Array, "n_wavelengths"],
                                      observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the ST magnitude in a given filter passband.

    This function computes the ST magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the ST magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^2/Å]

    Returns:
        float: ST magnitude in the filter passband [mag]
    """
    transmission_responses = filter.filter_responses_for_wavelengths(wavelengths)
    
    return -2.5 * jnp.log10(
            trapezoid(x=wavelengths, y=wavelengths * observed_flux / 1e8 * transmission_responses) /
            trapezoid(x=wavelengths, y=wavelengths * transmission_responses)
        ) + filter.ST_zeropoint
    
    
def ST_passband_luminosity(filter: Filter,
                           wavelengths: Float[Array, "n_wavelengths"],
                           observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the ST magnitude in a given filter passband.

    This function computes the ST magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the ST magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^2/Å]

    Returns:
        float: ST magnitude in the filter passband [mag]
    """
    if 'gaia' in filter.name.lower():
        raise ValueError("Gaia filters are not supported for ST magnitude calculations.")
    elif 'panstarrs' in filter.name.lower():
        return __ST_passband_luminosity_non_photonic_panstarrs_ps1(filter, wavelengths, observed_flux)
    elif filter.non_photonic:
        return __ST_passband_luminosity_non_photonic(filter, wavelengths, observed_flux)
    else:
        return __ST_passband_luminosity_photonic(filter, wavelengths, observed_flux)


@partial(jax.jit, static_argnums=(0,))
def __Vega_passband_luminosity_photonic_gaia(filter: Filter,
                                             wavelengths: Float[Array, "n_wavelengths"],
                                             observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the Vega magnitude in a given filter passband.

    This function computes the Vega magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the Vega magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^2/Å]

    Returns:
        float: AB magnitude in the filter passband [mag]
    """
    transmission_responses = filter.filter_responses_for_wavelengths(wavelengths)
    # Montegriffo (2023)
    # P_A/(10^9 h*c) * int(lambda * f_lambda * T_zeta dlambda)
    
    nm_lambda = wavelengths/10.
    watt_flux = 1e-10*observed_flux

    return -2.5 * jnp.log10(3663830098681164.5* 
            trapezoid(x=nm_lambda, y=nm_lambda * watt_flux * transmission_responses)
    ) + filter.Vega_zeropoint

    
@partial(jax.jit, static_argnums=(0,))
def __Vega_passband_luminosity_non_photonic(filter: Filter,
                                            wavelengths: Float[Array, "n_wavelengths"],
                                            observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the Vega magnitude in a given filter passband.

    This function computes the Vega magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the Vega magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^3]

    Returns:
        float: Vega magnitude in the filter passband [mag]
    """
    transmission_responses = filter.filter_responses_for_wavelengths(wavelengths)
    
    return -2.5 * jnp.log10(
            trapezoid(x=wavelengths, y=observed_flux / 1e8 * transmission_responses) /
            trapezoid(x=wavelengths, y=transmission_responses)
        ) + filter.Vega_zeropoint

@partial(jax.jit, static_argnums=(0,))
def __Vega_passband_luminosity_photonic(filter: Filter,
                                        wavelengths: Float[Array, "n_wavelengths"],
                                        observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the Vega magnitude in a given filter passband.

    This function computes the Vega magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the Vega magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^3]

    Returns:
        float: Vega magnitude in the filter passband [mag]
    """
    transmission_responses = filter.filter_responses_for_wavelengths(wavelengths)
    
    return -2.5 * jnp.log10(
            trapezoid(x=wavelengths, y=wavelengths * observed_flux / 1e8 * transmission_responses) /
            trapezoid(x=wavelengths, y=wavelengths * transmission_responses)
        ) + filter.Vega_zeropoint


def Vega_passband_luminosity(filter: Filter,
                             wavelengths: Float[Array, "n_wavelengths"],
                             observed_flux: Float[Array, "n_wavelengths 2"]) -> float:
    """Calculate the Vega magnitude in a given filter passband.

    This function computes the Vega magnitude by integrating the observed flux weighted by
    the filter transmission function and comparing to the Vega magnitude zero point.

    Args:
        filter (Filter): Filter object containing the transmission curve
        wavelengths (Float[Array, "n_wavelengths"]): Wavelength points [Angstrom]
        observed_flux (Float[Array, "n_wavelengths 2"]): Observed flux at each wavelength point
            [erg/s/cm^2/Å]

    Returns:
        float: AB magnitude in the filter passband [mag]
    """
    if filter.Vega_zeropoint is None:
        raise ValueError("Vega zeropoint is not set for this filter.")
    
    if 'gaia' in filter.name.lower():
        return __Vega_passband_luminosity_photonic_gaia(filter, wavelengths, observed_flux)
    elif filter.non_photonic:
        return __Vega_passband_luminosity_non_photonic(filter, wavelengths, observed_flux)
    else:
        return __Vega_passband_luminosity_photonic(filter, wavelengths, observed_flux)


@jax.jit
def absolute_bol_luminosity(luminosity: float) -> float:
    """Calculate the absolute bolometric magnitude from a given luminosity.

    This function converts a bolometric luminosity in erg/s to an absolute bolometric
    magnitude using the standard formula M_bol = -2.5 * log10(L) + M_bol,⊙, where
    M_bol,⊙ = 71.1974 is the zero point calibrated to the Sun's bolometric magnitude.

    Args:
        luminosity (float): Total bolometric luminosity [erg/s]

    Returns:
        float: Absolute bolometric magnitude [mag]
    """
    return -2.5 * jnp.log10(luminosity * ERG_S_TO_W) + 71.1974



from .spectrum import simulate_observed_flux, simulate_monochromatic_luminosity, luminosity, AB_passband_luminosity, absolute_bol_luminosity
from .utils import ERG_S_TO_W, SPHERE_STERADIAN, ZERO_POINT_LUM_W
from .spectrum_emulator import SpectrumEmulator


from typing import Tuple

try:
    import jax
    import jax.numpy as jnp
    from jax.typing import ArrayLike
except ImportError:
    raise ImportError("Please install JAX to use SPICE.")


ERG_S_TO_W = 1e-7
SPHERE_STERADIAN = 12.56
ZERO_POINT_LUM_W = 3.0128*10**(28)
C_CENTIMETERS = 29979245800.
JY_TO_ERG = 1e-23
H_CONST_ERG_S = 6.62607015*10**(-27)


@jax.jit
def wavelengths_to_frequencies(wavelengths: ArrayLike) -> ArrayLike:
    return C_CENTIMETERS*1e8/wavelengths


@jax.jit
def intensity_wavelengths_to_hz(wavelengths: ArrayLike, intensity: ArrayLike) -> Tuple[ArrayLike, ArrayLike]:
    vws_hz = wavelengths_to_frequencies(wavelengths)
    fv_Jy = (intensity*1e-4*3.34*wavelengths**2)
    
    vws_mask = jnp.argsort(vws_hz)
    fv_Jy = fv_Jy[vws_mask]
    vws_hz = jnp.sort(vws_hz)
    
    return vws_hz, fv_Jy


@jax.jit
def intensity_Jy_to_erg(intensity: ArrayLike) -> ArrayLike:
    return intensity*JY_TO_ERG

def linear_interpolation_1d(x: jnp.ndarray, 
                          y: jnp.ndarray, 
                          x_query: jnp.ndarray) -> jnp.ndarray:
    """
    1D linear interpolation using JAX.
    
    Args:
        x: Array of x coordinates (must be sorted)
        y: Array of y coordinates 
        x_query: Points to interpolate at
        
    Returns:
        Interpolated values at x_query points
    """
    # Find indices of closest points
    idx = jnp.searchsorted(x, x_query) - 1
    idx = jnp.clip(idx, 0, len(x)-2)
    
    # Get surrounding points
    x0 = x[idx]
    x1 = x[idx + 1]
    y0 = y[idx]
    y1 = y[idx + 1]
    
    # Linear interpolation
    t = (x_query - x0) / (x1 - x0)
    return y0 + t * (y1 - y0)

@jax.jit
def linear_multivariate_interpolation(points: jnp.ndarray,
                                      values: jnp.ndarray,
                                      query_points: jnp.ndarray) -> jnp.ndarray:
    """
    Multivariate interpolation using JAX.
    
    Args:
        points: Array of points (N x D)
        values: Array of values at points (N,)
        query_points: Points to interpolate at (M x D)
        
    Returns:
        Interpolated values at query points (M,)
    """
    # Compute weights using inverse distance weighting
    distances = jnp.sqrt(jnp.sum((points[:, None, :] - query_points[None, :, :]) ** 2, axis=-1))
    weights = 1.0 / (distances + 1e-10)
    weights = weights / jnp.sum(weights, axis=0, keepdims=True)
    
    # Compute weighted average
    return jnp.sum(weights * values[:, None], axis=0)


@jax.jit
def nearest_multivariate_interpolation(points: jnp.ndarray,
                                       values: jnp.ndarray,
                                       query_points: jnp.ndarray) -> jnp.ndarray:
    """
    Multivariate interpolation using JAX.
    
    Args:
        points: Array of points (N x D)
        values: Array of values at points (N,)
        query_points: Points to interpolate at (M x D)
        
    Returns:
        Interpolated values at query points (M,)
    """
    # Compute pairwise distances
    distances = jnp.sum((points[:, None, :] - query_points[None, :, :]) ** 2, axis=-1)
    # Find nearest neighbor
    nearest_idx = jnp.argmin(distances, axis=0)
    return values[nearest_idx]



import pickle
import jax
import numpy as np
from spice.spectrum.spectrum_emulator import SpectrumEmulator
from overrides import override
from typing import List, Optional
from numpy.typing import ArrayLike
import jax.numpy as jnp
from huggingface_hub import hf_hub_download
import h5py

from spice.spectrum.utils import linear_multivariate_interpolation

METALLICITY_NAMES = ["default_metals_h", "metals/h", "me/h",
                      "[metals/h]", "[me/h]", "metallicity", "metals_h", "me_h"]

LOG_G_NAMES: List[str] = ['logg', 'log_g', 'log g',
                          'surface gravity',
                          'surface_gravity']
TEFF_NAMES: List[str] = ['teff', 't_eff', 't eff',
                         'effective_temperature',
                         'effective temperature',]

label_names = ["teff", "logg", "m/h"]

REPO_ID = "mjablonska/spice_korg_interpolator"
REGULAR_FILENAME = "regular_grid.h5"
SMALL_FILENAME = "small_grid.h5"
DEFAULT_CACHE_PATH = '~/.spice_cache'


@jax.jit
def _interpolate_spectrum(log10_wavelengths, continuum_wavelengths, intensities, continuum_intensities, all_parameters, parameters, log10_wavelength):
    wave_idx = jnp.searchsorted(log10_wavelengths, log10_wavelength)
    wave_indices = jnp.clip(jnp.array([wave_idx - 1, wave_idx]), 0, len(log10_wavelengths) - 1)
    
    continuum_wave_idx = jnp.searchsorted(continuum_wavelengths, log10_wavelength)
    continuum_wave_indices = jnp.clip(jnp.array([continuum_wave_idx - 1, continuum_wave_idx]), 0, len(continuum_wavelengths) - 1)
    
    repeated_params = jnp.repeat(all_parameters, 2, axis=0)
    repeated_log10_wavelengths = jnp.tile(log10_wavelengths[wave_indices],
                                            (all_parameters.shape[0], 1)).reshape((-1, 1))
    repeated_continuum_wavelengths = jnp.tile(continuum_wavelengths[continuum_wave_indices],
                                            (all_parameters.shape[0], 1)).reshape((-1, 1))
    
    params_with_wavelength = jnp.hstack([parameters, log10_wavelength]).reshape(1, -1)
    
    return jnp.concatenate([
        linear_multivariate_interpolation(
            jnp.hstack([repeated_params, repeated_log10_wavelengths]),
            intensities[:, [wave_indices]].flatten(),
            params_with_wavelength
        ),
        linear_multivariate_interpolation(
            jnp.hstack([repeated_params, repeated_continuum_wavelengths]),
            continuum_intensities[:, [continuum_wave_indices]].flatten(),
            params_with_wavelength
        )
    ])

interpolate_spectrum = jax.vmap(_interpolate_spectrum, in_axes=(None, None, None, None, None, None, 0))


class KorgSpectrumEmulator(SpectrumEmulator[ArrayLike]):
    def __init__(self, cache_path: str = DEFAULT_CACHE_PATH, model_path: Optional[str] = None, grid_type: str = "small"):
        if model_path is not None:
            try:
                model_path = pickle.load(model_path)
            except Exception as e:
                raise RuntimeError(f"Failed to load model from path: {e}. If no model_path is provided, the model will be downloaded from the Hugging Face Hub.")
        
        else:
            print("Using default cache path:", cache_path)
            self.cache_path = cache_path
            # Check if the file exists in cache
            try:
                print("Attempting to download and load model from Hugging Face Hub...")
                filename = SMALL_FILENAME if grid_type == "small" else REGULAR_FILENAME
                model_path = hf_hub_download(repo_id=REPO_ID, filename=filename)
                print("Model downloaded to:", model_path)
                
                with h5py.File(model_path, 'r') as f:
                    print("Loading model parameters...")
                    self.parameters = jnp.array(f['parameters']) # [teff, logg, metallicity, mu]
                    print("Parameters shape:", self.parameters.shape)
                    
                    print("Loading model intensities...")
                    self.intensities = jnp.array(f['sp_intensity'])
                    print("Intensities shape:", self.intensities.shape)
                    
                    print("Loading wavelengths...")
                    self.log10_wavelengths = jnp.array(f['log10_sp_wave'])
                    print("Wavelengths shape:", self.log10_wavelengths.shape)
                    
                    print("Loading continuum intensities...")
                    self.continuum_intensities = jnp.array(f['sp_no_lines_intensity'])
                    print("Continuum intensities shape:", self.continuum_intensities.shape)
                    
                    print("Loading continuum wavelengths...")
                    self.continuum_wavelengths = jnp.array(f['log10_sp_no_lines_wave'])
                    print("Continuum wavelengths shape:", self.continuum_wavelengths.shape)
                
                print("Model loaded successfully")
            except Exception as e:
                print("Error loading model:", str(e))
                raise RuntimeError(f"Failed to load model from cache: {e}")

    @property
    def parameter_names(self) -> ArrayLike:
        return label_names

    @property
    @override
    def stellar_parameter_names(self) -> ArrayLike:
        return label_names

    @override
    def to_parameters(self, parameters: ArrayLike = None) -> ArrayLike:
        parameters = parameters or {}
        if isinstance(parameters, (list, tuple)):
            parameters = jnp.array(parameters)
        if isinstance(parameters, np.ndarray):
            if len(parameters.shape) > 1:
                if parameters.shape[1] != len(label_names):
                    raise ValueError(
                        f"Parameters must have {len(label_names)} columns, got {parameters.shape[1]}")
                return parameters
            return parameters

        elif isinstance(parameters, dict):
            # Extract Teff from various possible keys
            teff = None
            for key in TEFF_NAMES:
                if key in parameters:
                    teff = parameters[key]
                    break
            if teff is None:
                teff = 5777.0
            
            logg = None
            for key in LOG_G_NAMES:
                if key in parameters:
                    logg = parameters[key]
                    break
            if logg is None:
                logg = 4.44
            
            mh = None
            for key in METALLICITY_NAMES:
                if key in parameters:
                    mh = parameters[key]
                    break
            if mh is None:
                mh = 0.0
            
            return jnp.array([teff, logg, mh])
        else:
            raise ValueError(
                f"Parameters must be an array, list, or dictionary, got {type(parameters)}")

    @override
    def flux(self, log_wavelengths: ArrayLike, parameters: ArrayLike) -> ArrayLike:
        raise NotImplementedError("KorgSpectrumEmulator does not support flux")

    @override
    def intensity(self, log_wavelengths: ArrayLike, mu: float, parameters: ArrayLike) -> ArrayLike:
        return interpolate_spectrum(self.log10_wavelengths, self.continuum_wavelengths, self.intensities, self.continuum_intensities, self.parameters,
                                    jnp.hstack([parameters, jnp.atleast_1d(mu)]), log_wavelengths)



"""Utilities for parameter modification and transformation."""

from typing import Any, Callable, Optional, Union, TypeVar
import jax
import jax.numpy as jnp
from jax.typing import ArrayLike
from ..models.mesh_model import MeshModel

T = TypeVar('T')


def modify_mesh_parameter(
    mesh: MeshModel,
    parameter_index: int,
    distribution_fn: Callable[[MeshModel, Optional[ArrayLike]], ArrayLike],
    weights: Optional[ArrayLike] = None,
    bounds: Optional[tuple[float, float]] = None,
    clip: bool = True
) -> MeshModel:
    """Modify mesh parameters according to a distribution function.
    
    This function applies a distribution function to modify parameters across a mesh,
    optionally weighted by some mesh property (e.g. area, distance from center, etc.).
    The function returns a new mesh with modified parameters.
    
    Args:
        mesh: The MeshModel instance to modify.
        parameter_index: Index of the parameter to modify in the mesh's parameters array.
        distribution_fn: Function that takes the parameter values and optional weights and returns modified values.
            Should be compatible with JAX transformations.
        weights: Optional weights to apply to the distribution, shape (n_elements,).
            If provided, the distribution will be weighted by these values.
        bounds: Optional tuple of (min_value, max_value) to constrain the parameters.
            If None, no bounds are applied.
        clip: If True and bounds are provided, clip values to the bounds.
            If False and bounds are provided, raise ValueError if values are out of bounds.
            
    Returns:
        Modified MeshModel with updated parameter values.
        
    Example:
        >>> # Create a temperature distribution that falls off with radius
        >>> def temp_distribution(temps, radii):
        ...     return temps * jnp.exp(-radii / scale_height)
        >>> # Get radii from mesh centers
        >>> radii = jnp.linalg.norm(mesh.d_centers, axis=1)
        >>> # Modify temperature parameter (assuming index 0)
        >>> new_mesh = modify_mesh_parameter(
        ...     mesh,
        ...     parameter_index=0,
        ...     distribution_fn=temp_distribution,
        ...     weights=radii
        ... )
    """
    # Apply distribution function
    if weights is not None:
        modified_values = distribution_fn(mesh, *weights)
    else:
        modified_values = distribution_fn(mesh)

    return _modify_mesh_parameter_impl(mesh, parameter_index, modified_values)


@jax.jit
def _modify_mesh_parameter_impl(
    mesh: MeshModel,
    parameter_index: int,
    modified_values: ArrayLike
) -> MeshModel:
    """Implementation of modify_mesh_parameter that can be JIT-compiled."""
    # Get current parameter values
    new_parameters = mesh.parameters.at[:, parameter_index].set(modified_values)
    
    return mesh._replace(parameters=new_parameters)


@jax.jit
def constant_luminosity_temperature(mp, original_radius, original_teff):
    """Calculate effective temperature that maintains constant luminosity as radius changes.
    It assumes a uniform temperature distribution.
    
    This function implements the temperature-radius relationship for constant luminosity:
    T ∝ R^(-1/2), derived from L ∝ R²T⁴ = constant.
    
    Args:
        mp: MeshModel or similar object containing radii information.
        original_radius: The reference radius value.
        original_teff: The effective temperature at the reference radius.
        
    Returns:
        float: The new effective temperature that maintains constant luminosity
              given the current radius of the model.
              
    Example:
        >>> # Calculate new temperature for a pulsating star
        >>> new_temp = constant_luminosity_temperature(
        ...     mesh_model, 
        ...     original_radius=1.0,  # Solar radii
        ...     original_teff=5800.0  # Kelvin
        ... )
    """
    
    # Get the current radius (which varies due to pulsation)
    current_radius = mp.radii.mean()  # Taking mean radius as representative
    
    # Calculate the new temperature using T ∝ R^(-1/2) relation
    # T_new = T_original * (R_original/R_new)^(1/2)
    scaling_factor = (original_radius / current_radius) ** 0.5
    new_teff = original_teff * scaling_factor
    return new_teff



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

#This is the script which generates the Roche's mesh for DI Her eclipsing eccentric binary.
#The orbital and physical parameters of DI Her were adopted after Albrecht et al. (2009).
#Tested with phoebe version 2.4.5

#------------------------------------------------------------------------
# MODULES
#------------------------------------------------------------------------
import phoebe
import numpy as np
from phoebe import u
import matplotlib.pyplot as plt

#------------------------------------------------------------------------
# INITIALIZATION OF LOGGER AND BINARY SYSTEM
#------------------------------------------------------------------------
logger = phoebe.logger(clevel='error',flevel='warning',filename='DI-Her.log')
b      = phoebe.default_binary()

#------------------------------------------------------------------------
# PARAMETERS OF THE SYSTEM
#------------------------------------------------------------------------
P  = 10.550164 #orbital period in days
T0 = 2442233.3481 #time of periastron passage in HJD

b.set_value('period', component='binary', value=P*u.d)
b.set_value('ecc', component='binary', value=0.489)
b.set_value('per0', component='binary', value=330.2*u.deg)
b.set_value('incl', component='binary', value=89.3*u.deg)
b.set_value('t0_supconj', component='binary', value=T0*u.d)
b.set_value('vgamma',value=9.1*u.km/u.s)

#------------------------------------------------------------------------
# PHYSICAL PARAMETERS OF THE COMPONENTS
#------------------------------------------------------------------------
b.set_value('teff', component='primary', value=17300*u.K)
b.set_value('teff', component='secondary', value=15400*u.K)

b.set_value('requiv', component='primary', value=2.68*u.solRad)
b.set_value('requiv', component='secondary', value=2.48*u.solRad)

b.set_value('gravb_bol', component='primary', value=1.0)
b.set_value('gravb_bol', component='secondary', value=1.0)

b.set_value('irrad_frac_refl_bol', component='primary', value=1.0)
b.set_value('irrad_frac_refl_bol', component='secondary', value=1.0)

#Synchronicity parameters have to be calculated from vsini for both components
vsini_p = 109.0                              #km/s
vsini_s = 117.0                              #km/s
sini    = np.sin(89.3*np.pi/180.0)
vp, vs  = vsini_p/sini, vsini_s/sini         #km/s
R_sun   = 695700000                          #m
Prot_p  = 2*np.pi*2.68*R_sun/(vp*1000)/86400 #d
Prot_s  = 2*np.pi*2.48*R_sun/(vs*1000)/86400 #d

b.set_value('syncpar', component='primary', value=P/Prot_p)
b.set_value('syncpar', component='secondary', value=P/Prot_s)

b.flip_constraint('mass@primary', solve_for='sma@binary')
b.set_value('mass', component='primary', value=5.1*u.solMass)
b.flip_constraint('mass@secondary', solve_for='q')
b.set_value('mass', component='secondary', value=4.4*u.solMass)

b.set_value('pitch', component='primary', value=-27.3*u.deg)
b.set_value('yaw', component='primary', value=72.0*u.deg)
b.set_value('pitch', component='secondary', value=10.7*u.deg)
b.set_value('yaw', component='secondary', value=-84.0*u.deg)

#------------------------------------------------------------------------
# NUMBER OF TRIANGLES
#------------------------------------------------------------------------
b['ntriangles@primary']   = 2500
b['ntriangles@secondary'] = 2500

#------------------------------------------------------------------------
# TIME GRID
#------------------------------------------------------------------------
phases = np.linspace(-0.07,0.07,25) #only 25 moments of time in the given range of orbital phases, just for the test
times  = b.to_times(phases)

#------------------------------------------------------------------------
# ATMOSPHERIC MODELS
#------------------------------------------------------------------------
b.set_value('atm',component='primary',value='ck2004')
b.set_value('atm',component='secondary',value='ck2004')
b.set_value('abun',component='primary',value=0.0)
b.set_value('abun',component='secondary',value=0.0)

#------------------------------------------------------------------------
# DATASETS TO COMPUTE
#------------------------------------------------------------------------
b.add_dataset('mesh',compute_times=times,dataset='bigmesh',coordinates='uvw',columns=['us','vs','ws','vws','areas','loggs','teffs','mus','visibilities'])

#Phoebe provides a lot of information for each surface element, I have extracted only fraction of them
#as they are necessary to integrate the composite spectrum:
#us,vs,ws – the positions of elements in plane of the sky [R_sun], u and v are co-planar with the sky and w is perpendicular to the sky
#vws – the radial velocities of each element [km/s]
#areas, loggs, teffs, mus – self-explanatory
#visibilities – their values indicate if the specific element is visible for an observer

#------------------------------------------------------------------------
# RUNNING COMPUTE
#------------------------------------------------------------------------
b.run_compute(irrad_method='horvat',distortion_method='roche',ltte=False,rv_method='flux-weighted',rv_grav=False)

#animation of DI Her + distribution of radial velocities and teffs during periastron passage
#just for check/fun :)
b['bigmesh@latest@model'].plot(fc='vws', ec='None', fclim=[-200,200], exclude_back=True, fcmap='seismic',axpos=211)
b['bigmesh@latest@model'].plot(fc='teffs', ec='None', fclim=[17100,17500], exclude_back=True, fcmap='plasma',axpos=212)
fig = plt.figure(figsize=(8,9),dpi=200)
b.savefig('DI_Her-vrad+teff.mp4', fig=fig, tight_layouot=True, draw_sidebars=True, animate=True, save_kwargs={'writer': 'ffmpeg'})

#------------------------------------------------------------------------
# ACCESSING VALUES FROM MESH FOR THE SPECIFIC MOMENT IN TIME
#------------------------------------------------------------------------
time = times[0]
U    = b.get_parameter(qualifier='us',component='primary',dataset='bigmesh',kind='mesh',time=time).value
V    = b.get_parameter(qualifier='vs',component='primary',dataset='bigmesh',kind='mesh',time=time).value
VW   = b.get_parameter(qualifier='vws',component='primary',dataset='bigmesh',kind='mesh',time=time).value

#example plot of the radial velocities of centers of triangular surface elements for the primary component
#just to show/check that everything works fine and smoothly :)
plt.close()
plt.figure(figsize=(8,6))
plt.gca().set_aspect('equal')
plt.scatter(U,V,marker='o',c=VW,s=7)
plt.xlabel(r'$u$ ($R_\odot$)',fontsize=14)
plt.ylabel(r'$v$ ($R_\odot$)',fontsize=14)
plt.title(r'Distribution of $v_{\rm rad}$ for the primary at $t=$'+str(np.around(time,4))+' HJD',fontsize=12)
plt.colorbar(label=r'$v_{\rm rad}$ (km/s)')
plt.tight_layout()
plt.show()



from .model import Model
from .mesh_generation import icosphere
from .mesh_model import IcosphereModel, MeshModel
from .spots import add_spherical_harmonic_spot, add_spherical_harmonic_spots
from .binary import Binary
from .phoebe_model import PhoebeModel
from .utils import lat_to_theta, lon_to_phi, theta_to_lat, phi_to_lon


from typing import TypeVar


T = TypeVar('T')


class Model:
    @property
    def vertices(self) -> T:
        raise NotImplementedError
    
    @property
    def mesh_elements(self) -> T:
        raise NotImplementedError
    
    @property
    def centers(self) -> T:
        raise NotImplementedError
        
    @property
    def velocities(self) -> T:
        raise NotImplementedError
    
    @property
    def mus(self) -> T:
        raise NotImplementedError
    
    @property
    def los_velocities(self) -> T:
        raise NotImplementedError
    
    @property
    def los_z(self) -> T:
        raise NotImplementedError

    @property
    def radii(self) -> T:
        raise NotImplementedError
    
    @property
    def cast_vertices(self) -> T:
        raise NotImplementedError
    
    @property
    def cast_centers(self) -> T:
        raise NotImplementedError

    @property
    def cast_areas(self) -> T:
        raise NotImplementedError

    @classmethod
    def construct(cls, **kwargs) -> "Model":
        raise NotImplementedError
